# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""FederatedDataset module."""

import numpy as np

from openfl.utilities.data_splitters import EqualNumPyDataSplitter
from openfl.utilities.data_splitters import NumPyDataSplitter
from .loader_pt import PyTorchDataLoader


class FederatedDataSet(PyTorchDataLoader):
    """
    Data Loader for in memory Numpy data.

    Args:
        X_train: np.array
            Training Features
        y_train: np.array
            Training labels
        X_val: np.array
            Validation features
        y_val: np.array
            Validation labels
        batch_size : int
            The batch size for the data loader
        num_classes : int
            The number of classes the model will be trained on
        **kwargs: Additional arguments to pass to the function

    """

    train_splitter: NumPyDataSplitter
    valid_splitter: NumPyDataSplitter

    def __init__(self, X_train, y_train, X_valid, y_valid,
                 batch_size=1, num_classes=None, train_splitter=None, valid_splitter=None):
        """
        Initialize.

        Args:
            X_train: np.array
                Training Features
            y_train: np.array
                Training labels
            X_val: np.array
                Validation features
            y_val: np.array
                Validation labels
            batch_size : int
                The batch size for the data loader
            num_classes : int
                The number of classes the model will be trained on
            train_splitter: NumPyDataSplitter
                Data splitter for train dataset.
            valid_splitter: NumPyDataSplitter
                Data splitter for validation dataset.
            **kwargs: Additional arguments to pass to the function

        """
        super().__init__(batch_size)

        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid

        if num_classes is None:
            num_classes = np.unique(self.y_train).shape[0]
            print(f'Inferred {num_classes} classes from the provided labels...')
        self.num_classes = num_classes
        self.train_splitter = self._get_splitter_or_default(train_splitter)
        self.valid_splitter = self._get_splitter_or_default(valid_splitter)

    @staticmethod
    def _get_splitter_or_default(value):
        if value is None:
            return EqualNumPyDataSplitter()
        if isinstance(value, NumPyDataSplitter):
            return value
        else:
            raise NotImplementedError(f'Data splitter {value} is not supported')

    def split(self, num_collaborators):
        """Create a Federated Dataset for each of the collaborators.

        Args:
            num_collaborators: int
                Collaborators to split the dataset between
            shuffle: boolean
                Should the dataset be randomized?
            equally: boolean
                Should each collaborator get the same amount of data?

        Returns:
            list[FederatedDataSets]
                A dataset slice for each collaborator
        """
        train_idx = self.train_splitter.split(self.y_train, num_collaborators)
        valid_idx = self.valid_splitter.split(self.y_valid, num_collaborators)

        return [
            FederatedDataSet(
                self.X_train[train_idx[i]],
                self.y_train[train_idx[i]],
                self.X_valid[valid_idx[i]],
                self.y_valid[valid_idx[i]],
                batch_size=self.batch_size,
                num_classes=self.num_classes,
                train_splitter=self.train_splitter,
                valid_splitter=self.valid_splitter
            ) for i in range(num_collaborators)
        ]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Percentage based Straggler Handling function."""
from openfl.component.straggler_handling_functions import StragglerHandlingFunction


class PercentageBasedStragglerHandling(StragglerHandlingFunction):
    def __init__(
        self,
        percent_collaborators_needed=1.0,
        minimum_reporting=1,
        **kwargs):
        self.percent_collaborators_needed = percent_collaborators_needed
        self.minimum_reporting = minimum_reporting

    def minimum_collaborators_reported(self, num_collaborators_done):
        return num_collaborators_done >= self.minimum_reporting

    def straggler_cutoff_check(self, num_collaborators_done, all_collaborators):
        cutoff = (num_collaborators_done >= self.percent_collaborators_needed * len(
            all_collaborators)) and self.minimum_collaborators_reported(num_collaborators_done)
        return cutoff



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import os

import nibabel as nib
import numpy as np
import tensorflow as tf


class DatasetGenerator:
    """Generate a TensorFlow data loader from the BraTS .nii.gz files."""

    def __init__(self, crop_dim,
                 data_path,
                 batch_size=4,
                 number_input_channels=1,
                 num_classes=1,
                 train_test_split=0.80,
                 validate_test_split=0.5,
                 random_seed=816,
                 shard=0):
        """Initialize the class."""
        self.data_path = os.path.abspath(os.path.expanduser(data_path))
        self.batch_size = batch_size
        self.crop_dim = [crop_dim, crop_dim, crop_dim, number_input_channels]
        self.num_input_channels = number_input_channels
        self.num_classes = num_classes
        self.random_seed = random_seed

        self.train_test_split = train_test_split
        self.validate_test_split = validate_test_split
        self.shard = shard

        self.create_file_list()

        if self.num_files > 0:

            self.ds_train, self.ds_val, self.ds_test = self.get_dataset()

        else:

            self.ds_train = None
            self.ds_val = None
            self.ds_test = None
            raise ValueError(f'ERROR: No BraTS datafiles found under directory {self.data_path}')

    def create_file_list(self):
        """
        Get list of the files from the BraTS raw data.

        Split into training and testing sets.
        """
        searchpath = os.path.join(self.data_path, '*/*_seg.nii.gz')
        filenames = tf.io.gfile.glob(searchpath)

        # Create a dictionary of tuples with image filename and label filename

        self.num_files = len(filenames)
        self.filenames = {}
        for idx, filename in enumerate(filenames):
            self.filenames[idx] = [filename.replace('_seg.nii.gz', '_flair.nii.gz'), filename]

    def z_normalize_img(self, img):
        """
        Normalize the image.

        The mean value for each image is 0 and the standard deviation is 1.
        """
        # TODO: Correct this for multiple MRI channels
        return (img - np.mean(img)) / np.std(img)

    def crop(self, img, msk, randomize):
        """Randomly crop the image and mask."""
        slices = []

        # Do we randomize?
        is_random = randomize and np.random.rand() > 0.5

        for idx in range(len(img.shape) - 1):  # Go through each dimension

            croplen = self.crop_dim[idx]
            imglen = img.shape[idx]

            start = (imglen - croplen) // 2

            ratio_crop = 0.20  # Crop up this this % of pixels for offset
            # Number of pixels to offset crop in this dimension
            offset = int(np.floor(start * ratio_crop))

            if offset > 0:
                if is_random:
                    start += np.random.choice(range(-offset, offset))
                    if ((start + croplen) > imglen):  # Don't fall off the image
                        start = (imglen - croplen) // 2
            else:
                start = 0

            slices.append(slice(start, start + croplen))

        return img[tuple(slices)], msk[tuple(slices)]

    def augment_data(self, img, msk):
        """
        Augmentation the input images.

        Flip image and mask. Rotate image and mask.
        """
        # Determine if axes are equal and can be rotated
        # If the axes aren't equal then we can't rotate them.
        equal_dim_axis = []
        for idx in range(0, len(self.crop_dim)):
            for jdx in range(idx + 1, len(self.crop_dim)):
                if self.crop_dim[idx] == self.crop_dim[jdx]:
                    equal_dim_axis.append([idx, jdx])  # Valid rotation axes
        dim_to_rotate = equal_dim_axis

        if np.random.rand() > 0.5:
            # Random 0,1 (axes to flip)
            ax = np.random.choice(np.arange(len(self.crop_dim) - 1))
            img = np.flip(img, ax)
            msk = np.flip(msk, ax)

        elif (len(dim_to_rotate) > 0) and (np.random.rand() > 0.5):
            rot = np.random.choice([1, 2, 3])  # 90, 180, or 270 degrees

            # This will choose the axes to rotate
            # Axes must be equal in size
            random_axis = dim_to_rotate[np.random.choice(len(dim_to_rotate))]

            img = np.rot90(img, rot, axes=random_axis)  # Rotate axes 0 and 1
            msk = np.rot90(msk, rot, axes=random_axis)  # Rotate axes 0 and 1

        return img, msk

    def read_nifti_file(self, idx, randomize=False):
        """Read Nifti file."""
        idx = idx.numpy()
        imgfile = self.filenames[idx][0]
        mskfile = self.filenames[idx][1]

        img_temp = np.array(nib.load(imgfile).dataobj)
        img_temp = np.rot90(img_temp)

        img = np.zeros(list(img_temp.shape) + [self.num_input_channels])
        # Normalize
        img_temp = self.z_normalize_img(img_temp)

        img[..., 0] = img_temp

        for channel in range(1, self.num_input_channels):

            if channel == 1:
                imgfile = self.filenames[idx][1].replace('_flair', '_t1')
            elif channel == 2:
                imgfile = self.filenames[idx][1].replace('_flair', '_t1ce')
            elif channel == 3:
                imgfile = self.filenames[idx][1].replace('_flair', '_t2')

            img_temp = np.array(nib.load(imgfile).dataobj)

            img_temp = np.rot90(img_temp)

            # Normalize
            img_temp = self.z_normalize_img(img_temp)

            img[..., channel] = img_temp

        msk = np.rot90(np.array(nib.load(mskfile).dataobj))
        msk = np.expand_dims(msk, -1)

        # labels: {
        #      0: background,
        #      1: edema,
        #      2: non-enhancing tumor,
        #      3: enhancing tumour}

        # Combine all masks but background
        if self.num_classes == 1:
            msk[msk > 0] = 1.0
        else:
            msk_temp = np.zeros(list(msk.shape) + [self.num_classes])
            for channel in range(self.num_classes):
                msk_temp[msk == channel, channel] = 1.0
            msk = msk_temp

        # Crop
        img, msk = self.crop(img, msk, randomize)

        # Randomly rotate
        if randomize:
            img, msk = self.augment_data(img, msk)

        return img, msk

    def get_input_shape(self):
        """Get the shape of the input."""
        return self.crop_dim

    def plot_images(self, ds, slice_num=90):
        """Plot images from dataset."""
        import matplotlib.pyplot as plt

        plt.figure(figsize=(20, 20))

        num_cols = 2

        msk_channel = 0
        img_channel = 0

        for img, msk in ds.take(1):
            bs = img.shape[0]

            for idx in range(bs):
                plt.subplot(bs, num_cols, idx * num_cols + 1)
                plt.imshow(img[idx, :, :, slice_num, img_channel], cmap='bone')
                plt.title('MRI', fontsize=18)
                plt.subplot(bs, num_cols, idx * num_cols + 2)
                plt.imshow(msk[idx, :, :, slice_num, msk_channel], cmap='bone')
                plt.title('Tumor', fontsize=18)

        plt.show()

        print(f'Mean pixel value of image = {np.mean(img[0, :, :, :, 0])}')

    def display_train_images(self, slice_num=90):
        """Plot some training images."""
        self.plot_images(self.ds_train, slice_num)

    def display_validation_images(self, slice_num=90):
        """Plot some validation images."""
        self.plot_images(self.ds_val, slice_num)

    def display_test_images(self, slice_num=90):
        """Plot some test images."""
        self.plot_images(self.ds_test, slice_num)

    def get_train(self):
        """Return train dataset."""
        return self.ds_train

    def get_test(self):
        """Return test dataset."""
        return self.ds_test

    def get_validate(self):
        """Return validation dataset."""
        return self.ds_val

    def get_dataset(self):
        """Create a TensorFlow data loader."""
        self.num_train = int(self.num_files * self.train_test_split)
        numvaltest = self.num_files - self.num_train

        ds = tf.data.Dataset.range(self.num_files).shuffle(
            self.num_files, self.random_seed)  # Shuffle the dataset

        # Horovod Sharding
        # Here we are not actually dividing the dataset into shards
        # but instead just reshuffling the training dataset for every
        # shard. Then in the training loop we just go through the training
        # dataset but the number of steps is divided by the number of shards.
        ds_train = ds.take(self.num_train).shuffle(
            self.num_train, self.shard)  # Reshuffle based on shard
        ds_val_test = ds.skip(self.num_train)
        self.num_val = int(numvaltest * self.validate_test_split)
        self.num_test = self.num_train - self.num_val
        ds_val = ds_val_test.take(self.num_val)
        ds_test = ds_val_test.skip(self.num_val)

        ds_train = ds_train.map(lambda x: tf.py_function(self.read_nifti_file,
                                                         [x, True], [tf.float32, tf.float32]),
                                num_parallel_calls=tf.data.experimental.AUTOTUNE)
        ds_val = ds_val.map(lambda x: tf.py_function(self.read_nifti_file,
                                                     [x, False], [tf.float32, tf.float32]),
                            num_parallel_calls=tf.data.experimental.AUTOTUNE)
        ds_test = ds_test.map(lambda x: tf.py_function(self.read_nifti_file,
                                                       [x, False], [tf.float32, tf.float32]),
                              num_parallel_calls=tf.data.experimental.AUTOTUNE)

        ds_train = ds_train.batch(self.batch_size, drop_remainder=True)
        ds_train = ds_train.prefetch(tf.data.experimental.AUTOTUNE)

        batch_size_val = max(1, self.batch_size // 2)    # Could be any batch size you'd like
        ds_val = ds_val.batch(batch_size_val, drop_remainder=True)
        ds_val = ds_val.prefetch(tf.data.experimental.AUTOTUNE)

        batch_size_test = max(1, self.batch_size // 2)   # Could be any batch size you'd like
        ds_test = ds_test.batch(batch_size_test, drop_remainder=True)
        ds_test = ds_test.prefetch(tf.data.experimental.AUTOTUNE)

        return ds_train, ds_val, ds_test



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""
pynvml CUDA Device monitor plugin module.

Required package: pynvml
"""

import pynvml

from .cuda_device_monitor import CUDADeviceMonitor

pynvml.nvmlInit()


class PynvmlCUDADeviceMonitor(CUDADeviceMonitor):
    """CUDA Device monitor plugin using pynvml lib."""

    def __init__(self) -> None:
        """Initialize pynvml plugin."""
        super().__init__()

    def get_driver_version(self) -> str:
        """Get Nvidia driver version."""
        return pynvml.nvmlSystemGetDriverVersion().decode('utf-8')

    def get_device_memory_total(self, index: int) -> int:
        """Get total memory available on the device."""
        handle = pynvml.nvmlDeviceGetHandleByIndex(index)
        info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        return info.total

    def get_device_memory_utilized(self, index: int) -> int:
        """Get utilized memory on the device."""
        handle = pynvml.nvmlDeviceGetHandleByIndex(index)
        info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        return info.used

    def get_device_utilization(self, index: int) -> str:
        """
        Get device utilization.

        It is just a general method that returns a string that may be shown to the frontend user.
        """
        handle = pynvml.nvmlDeviceGetHandleByIndex(index)
        info_utilization = pynvml.nvmlDeviceGetUtilizationRates(handle)
        return f'{info_utilization.gpu}%'

    def get_device_name(self, index: int) -> str:
        """Get device utilization method."""
        handle = pynvml.nvmlDeviceGetHandleByIndex(index)
        device_name = pynvml.nvmlDeviceGetName(handle)
        return device_name

    def get_cuda_version(self) -> str:
        """
        Get CUDA driver version.

        The CUDA version is specified as (1000 * major + 10 * minor),
        so CUDA 11.2 should be specified as 11020.
        https://docs.nvidia.com/cuda/cuda-driver-api/group__CUDA__DRIVER__ENTRY__POINT.html
        """
        cuda_version = pynvml.nvmlSystemGetCudaDriverVersion()
        major_version = int(cuda_version / 1000)
        minor_version = int(cuda_version % 1000 / 10)
        return f'{major_version}.{minor_version}'



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import tensorflow as tf

from openfl.federated import KerasTaskRunner


class TensorFlowCNN(KerasTaskRunner):
    """Initialize.

    Args:
        **kwargs: Additional parameters to pass to the function

    """

    def __init__(self, **kwargs):
        """Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function

        """
        super().__init__(**kwargs)

        self.model = self.create_model(
            self.feature_shape,
            self.data_loader.num_classes,
            **kwargs
        )
        self.initialize_tensorkeys_for_functions()

    def create_model(self,
                     input_shape,
                     num_classes,
                     training_smoothing=32.0,
                     validation_smoothing=1.0,
                     **kwargs):
        """Create the TensorFlow CNN Histology model.

        Args:
            training_smoothing (float): (Default=32.0)
            validation_smoothing (float): (Default=1.0)
            **kwargs: Additional parameters to pass to the function

        """
        print(tf.config.threading.get_intra_op_parallelism_threads())
        print(tf.config.threading.get_inter_op_parallelism_threads())
        # ## Define Model
        #
        # Convolutional neural network model

        inputs = tf.keras.layers.Input(shape=input_shape)
        conv = tf.keras.layers.Conv2D(
            filters=16, kernel_size=(3, 3), padding='same', activation='relu')(inputs)
        conv = tf.keras.layers.Conv2D(
            filters=32, kernel_size=(3, 3), padding='same', activation='relu')(conv)
        maxpool = tf.keras.layers.MaxPooling2D(pool_size=(2, 2))(conv)

        conv = tf.keras.layers.Conv2D(
            filters=64, kernel_size=(3, 3), padding='same', activation='relu')(maxpool)
        conv = tf.keras.layers.Conv2D(
            filters=128, kernel_size=(3, 3), padding='same', activation='relu')(conv)
        concat = tf.keras.layers.concatenate([maxpool, conv])
        maxpool = tf.keras.layers.MaxPooling2D(pool_size=(2, 2))(concat)

        conv = tf.keras.layers.Conv2D(
            filters=256, kernel_size=(3, 3), padding='same', activation='relu')(maxpool)
        conv = tf.keras.layers.Conv2D(
            filters=512, kernel_size=(3, 3), padding='same', activation='relu')(conv)
        concat = tf.keras.layers.concatenate([maxpool, conv])
        maxpool = tf.keras.layers.MaxPooling2D(pool_size=(2, 2))(concat)

        conv = tf.keras.layers.Conv2D(
            filters=256, kernel_size=(3, 3), padding='same', activation='relu')(maxpool)
        conv = tf.keras.layers.Conv2D(
            filters=512, kernel_size=(3, 3), padding='same', activation='relu')(conv)
        concat = tf.keras.layers.concatenate([maxpool, conv])
        maxpool = tf.keras.layers.MaxPooling2D(pool_size=(2, 2))(concat)

        flat = tf.keras.layers.Flatten()(maxpool)
        dense = tf.keras.layers.Dense(128)(flat)
        drop = tf.keras.layers.Dropout(0.5)(dense)

        predict = tf.keras.layers.Dense(num_classes)(drop)

        model = tf.keras.models.Model(inputs=[inputs], outputs=[predict])

        self.optimizer = tf.keras.optimizers.Adam()

        model.compile(
            loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
            optimizer=self.optimizer,
            metrics=[tf.keras.metrics.SparseCategoricalAccuracy()],
        )

        self.tvars = model.layers
        print(f'layer names: {[var.name for var in self.tvars]}')

        self.opt_vars = self.optimizer.variables()
        print(f'optimizer vars: {self.opt_vars}')

        # Two opt_vars for one tvar: gradient and square sum for RMSprop.
        self.fl_vars = self.tvars + self.opt_vars

        return model



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Base abstract optimizer class module."""
import abc
from importlib import import_module
from os.path import splitext
from typing import Dict

from numpy import ndarray

from openfl.plugins.frameworks_adapters.framework_adapter_interface import (
    FrameworkAdapterPluginInterface
)


class Optimizer(abc.ABC):
    """Base abstract optimizer class."""

    @abc.abstractmethod
    def step(self, gradients: Dict[str, ndarray]) -> None:
        """Perform a single step for parameter update.

        Args:
            gradients: Partial derivatives with respect to optimized parameters.
        """
        pass

    def _set_params_from_model(self, model_interface):
        """Eject and store model parameters."""
        class_name = splitext(model_interface.framework_plugin)[1].strip('.')
        module_path = splitext(model_interface.framework_plugin)[0]
        framework_adapter = import_module(module_path)
        framework_adapter_plugin: FrameworkAdapterPluginInterface = getattr(
            framework_adapter, class_name, None)
        self.params: Dict[str, ndarray] = framework_adapter_plugin.get_tensor_dict(
            model_interface.provide_model())



"""Copyright (C) 2020-2021 Intel Corporation
   SPDX-License-Identifier: Apache-2.0

Licensed subject to the terms of the separately executed evaluation
license agreement between Intel Corporation and you.
"""
from logging import getLogger

import numpy as np
import src.dataloader_utils as dlu

from openfl.federated import KerasDataLoader

logger = getLogger(__name__)


class NLPDataLoader(KerasDataLoader):
    """NLP Dataloader template."""

    def __init__(self, collaborator_count, split_ratio,
                 num_samples, data_path, batch_size, **kwargs):
        """Instantiate the data object.

        Args:
            data_path: The file path to the data Returns:
            batch_size: The batch size of the data loader tuple: shape of an example feature array
            **kwargs: Additional arguments, passed to super init and load_mnist_shard

        Returns:
           none
        """
        self.shard_num = data_path
        self.data_path = dlu.download_data_()

        self.batch_size = batch_size

        train, valid, details = dlu.load_shard(collaborator_count, self.shard_num,
                                               self.data_path, num_samples, split_ratio)

        self.num_samples = details['num_samples']
        self.num_encoder_tokens = details['num_encoder_tokens']
        self.num_decoder_tokens = details['num_decoder_tokens']
        self.max_encoder_seq_length = details['max_encoder_seq_length']
        self.max_decoder_seq_length = details['max_decoder_seq_length']

        self.X_train = [train[0], train[1]]
        self.y_train = train[2]
        self.X_valid = [valid[0], valid[1]]
        self.y_valid = valid[2]

    def get_feature_shape(self):
        """Get the shape of an example feature array."""
        return self.X_train[0].shape

    def get_train_loader(self, batch_size=None):
        """
        Get training data loader.

        Returns
        -------
        loader object
        """
        return self._get_batch_generator(X1=self.X_train[0], X2=self.X_train[1],
                                         y=self.y_train, batch_size=batch_size)

    def get_valid_loader(self, batch_size=None):
        """
        Get validation data loader.

        Returns:
            loader object
        """
        return self._get_batch_generator(X1=self.X_valid[0], X2=self.X_valid[1],
                                         y=self.y_valid, batch_size=batch_size)

    def get_train_data_size(self):
        """
        Get total number of training samples.

        Returns:
            int: number of training samples
        """
        return self.X_train[0].shape[0]

    def get_valid_data_size(self):
        """
        Get total number of validation samples.

        Returns:
            int: number of validation samples
        """
        return self.X_valid[0].shape[0]

    @staticmethod
    def _batch_generator(X1, X2, y, idxs, batch_size, num_batches):
        """
        Generate batch of data.

        Args:
            X: input data
            y: label data
            idxs: The index of the dataset
            batch_size: The batch size for the data loader
            num_batches: The number of batches
        Yields:
            tuple: input data, label data
        """
        for i in range(num_batches):
            a = i * batch_size
            b = a + batch_size
            yield [X1[idxs[a:b]], X2[idxs[a:b]]], y[idxs[a:b]]

    def _get_batch_generator(self, X1, X2, y, batch_size):
        """
        Return the dataset generator.

        Args:
            X1: input data  (encoder)
            X2: input data  (decoder)
            y: label data
            batch_size: The batch size for the data loader
        """
        if batch_size is None:
            batch_size = self.batch_size
        # shuffle data indices
        idxs = np.random.permutation(np.arange(X1.shape[0]))
        # compute the number of batches
        num_batches = int(np.ceil(X1.shape[0] / batch_size))
        # build the generator and return it
        # TODO: due to _batch_generator(X1, ...) has first param X1, all params here will be moved,
        #       X1 -> X2, X2 -> y, y -> idxs, idxs -> batch_size, batch_size -> num_batches,
        #       and num_batches -> should be unexpected in this function
        return self._batch_generator(X1, X2, y, idxs, batch_size, num_batches)



from openfl.interface.aggregation_functions import AggregationFunction


class AggregateRandomForest(AggregationFunction):
    """Function for SciKit Learn Random Forest aggregation"""

    def call(self, local_tensors, *_):
        result = local_tensors[0].tensor
        for i in range(1, len(local_tensors)):
            model = local_tensors[i].tensor
            result.estimators_ += model.estimators_
        result.n_estimators = len(result.estimators_)

        return result



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Director CLI."""

import logging
import shutil
import sys
from pathlib import Path

import click
from click import group
from click import option
from click import pass_context
from click import Path as ClickPath
from dynaconf import Validator

from openfl.component.director import Director
from openfl.interface.cli_helper import WORKSPACE
from openfl.transport import DirectorGRPCServer
from openfl.utilities import merge_configs
from openfl.utilities.path_check import is_directory_traversal

logger = logging.getLogger(__name__)


@group()
@pass_context
def director(context):
    """Manage Federated Learning Director."""
    context.obj['group'] = 'director'


@director.command(name='start')
@option('-c', '--director-config-path', default='director.yaml',
        help='The director config file path', type=ClickPath(exists=True))
@option('--tls/--disable-tls', default=True,
        is_flag=True, help='Use TLS or not (By default TLS is enabled)')
@option('-rc', '--root-cert-path', 'root_certificate', required=False,
        type=ClickPath(exists=True), default=None,
        help='Path to a root CA cert')
@option('-pk', '--private-key-path', 'private_key', required=False,
        type=ClickPath(exists=True), default=None,
        help='Path to a private key')
@option('-oc', '--public-cert-path', 'certificate', required=False,
        type=ClickPath(exists=True), default=None,
        help='Path to a signed certificate')
@option('-nn', '--neural-network', 'nn', required=False,
        type=bool, default=True,
        help='Indicates if the models used are neural networks or not')
def start(director_config_path, tls, root_certificate, private_key, certificate, nn):
    """Start the director service."""
    director_config_path = Path(director_config_path).absolute()
    logger.info('🧿 Starting the Director Service.')
    if is_directory_traversal(director_config_path):
        click.echo('The director config file path is out of the openfl workspace scope.')
        sys.exit(1)
    config = merge_configs(
        settings_files=director_config_path,
        overwrite_dict={
            'root_certificate': root_certificate,
            'private_key': private_key,
            'certificate': certificate,
        },
        validators=[
            Validator('settings.listen_host', default='localhost'),
            Validator('settings.listen_port', default=50051, gte=1024, lte=65535),
            Validator('settings.sample_shape', default=[]),
            Validator('settings.target_shape', default=[]),
            Validator('settings.envoy_health_check_period', gte=1, lte=24 * 60 * 60),
        ],
        value_transform=[
            ('settings.sample_shape', lambda x: list(map(str, x))),
            ('settings.target_shape', lambda x: list(map(str, x))),
        ],
    )

    logger.info(
        f'Sample shape: {config.settings.sample_shape}, '
        f'target shape: {config.settings.target_shape}'
    )

    if config.root_certificate:
        config.root_certificate = Path(config.root_certificate).absolute()

    if config.private_key:
        config.private_key = Path(config.private_key).absolute()

    if config.certificate:
        config.certificate = Path(config.certificate).absolute()

    director_server = DirectorGRPCServer(
        director_cls=Director,
        tls=tls,
        sample_shape=config.settings.sample_shape,
        target_shape=config.settings.target_shape,
        root_certificate=config.root_certificate,
        private_key=config.private_key,
        certificate=config.certificate,
        settings=config.settings,
        listen_host=config.settings.listen_host,
        listen_port=config.settings.listen_port,
        nn=nn
    )
    director_server.start()


@director.command(name='create-workspace')
@option('-p', '--director-path', required=True,
        help='The director path', type=ClickPath())
def create(director_path):
    """Create a director workspace."""
    if is_directory_traversal(director_path):
        click.echo('The director path is out of the openfl workspace scope.')
        sys.exit(1)
    director_path = Path(director_path).absolute()
    if director_path.exists():
        if not click.confirm('Director workspace already exists. Recreate?', default=True):
            sys.exit(1)
        shutil.rmtree(director_path)
    (director_path / 'cert').mkdir(parents=True, exist_ok=True)
    (director_path / 'logs').mkdir(parents=True, exist_ok=True)
    shutil.copyfile(WORKSPACE / 'default/director.yaml', director_path / 'director.yaml')



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Cats and dogs shard descriptor."""

import json
import os
import shutil
from hashlib import md5
from logging import getLogger
from pathlib import Path
from random import shuffle
from typing import Optional
from zipfile import ZipFile

import numpy as np
from kaggle.api.kaggle_api_extended import KaggleApi
from PIL import Image

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor

logger = getLogger(__name__)


class DogsCatsShardDataset(ShardDataset):
    """Dogs and cats Shard dataset class."""

    def __init__(self, data_type: str, dataset_dir: Path,
                 rank: int = 1, worldsize: int = 1, enforce_image_hw=None):
        """Initialize DogsCatsShardDataset."""
        self.rank = rank
        self.worldsize = worldsize
        self.dataset_dir = dataset_dir
        self.enforce_image_hw = enforce_image_hw
        self.img_path = self.dataset_dir / data_type

        self.img_names = [
            img.name
            for img in sorted(self.img_path.iterdir())
            if img.suffix == '.jpg'
        ]

        # Sharding
        self.img_names = self.img_names[self.rank - 1::self.worldsize]
        # Shuffling the results dataset after choose half pictures of each class
        shuffle(self.img_names)

    def __getitem__(self, index):
        """Return a item by the index."""
        name = self.img_names[index]
        # Reading data
        img = Image.open(self.img_path / name)
        img_class = 1 if name[:3] == 'dog' else 0
        assert name[:3] in {'cat', 'dog'}, 'Wrong object classification'

        if self.enforce_image_hw is not None:
            # If we need to resize data
            # PIL accepts (w,h) tuple, not (h,w)
            img = img.resize(self.enforce_image_hw[::-1])

        img = np.asarray(img)

        assert img.shape[2] == 3

        return img, np.asarray([img_class], dtype=np.uint8)

    def __len__(self):
        """Return the len of the dataset."""
        return len(self.img_names)


class DogsCatsShardDescriptor(ShardDescriptor):
    """Shard descriptor class."""

    def __init__(self, data_folder: str = 'data',
                 rank_worldsize: str = '1,3',
                 enforce_image_hw: Optional[str] = None) -> None:
        """Initialize DogsCatsShardDescriptor."""
        super().__init__()
        # Settings for sharding the dataset
        self.rank, self.worldsize = map(int, rank_worldsize.split(','))

        self.data_folder = Path.cwd() / data_folder
        self.download_dataset()

        # Settings for resizing data
        self.enforce_image_hw = None
        if enforce_image_hw is not None:
            self.enforce_image_hw = tuple(map(int, enforce_image_hw.split(',')))

        # Calculating data and target shapes
        ds = self.get_dataset()
        sample, target = ds[0]
        self._sample_shape = [str(dim) for dim in sample.shape]
        self._target_shape = [str(*target.shape)]

        assert self._target_shape[0] == '1', 'Target shape Error'

    def download_dataset(self):
        """Download dataset from Kaggle."""
        if not os.path.exists(self.data_folder):
            os.mkdir(self.data_folder)

        if not self.is_dataset_complete():
            logger.info('Your dataset is absent or damaged. Downloading ... ')
            api = KaggleApi()
            api.authenticate()

            if os.path.exists('data/train'):
                shutil.rmtree('data/train')

            api.competition_download_file(
                'dogs-vs-cats-redux-kernels-edition',
                'train.zip', path=self.data_folder
            )

            with ZipFile(self.data_folder / 'train.zip', 'r') as zipobj:
                zipobj.extractall(self.data_folder)

            os.remove(self.data_folder / 'train.zip')

            self.save_all_md5()

    def get_dataset(self, dataset_type='train'):
        """Return a shard dataset by type."""
        return DogsCatsShardDataset(
            data_type=dataset_type,
            dataset_dir=self.data_folder,
            rank=self.rank,
            worldsize=self.worldsize,
            enforce_image_hw=self.enforce_image_hw
        )

    def calc_all_md5(self):
        """Calculate hash of all dataset."""
        md5_dict = {}
        for root, _, files in os.walk(self.data_folder):
            for file in files:
                if file == 'dataset.json':
                    continue
                md5_calc = md5(usedforsecurity=False)
                rel_dir = os.path.relpath(root, self.data_folder)
                rel_file = os.path.join(rel_dir, file)

                with open(self.data_folder / rel_file, 'rb') as f:
                    for chunk in iter(lambda: f.read(4096), b''):
                        md5_calc.update(chunk)
                    md5_dict[rel_file] = md5_calc.hexdigest()
        return md5_dict

    def save_all_md5(self):
        """Save dataset hash."""
        all_md5 = self.calc_all_md5()
        with open(os.path.join(self.data_folder, 'dataset.json'), 'w') as f:
            json.dump(all_md5, f)

    def is_dataset_complete(self):
        """Check dataset integrity."""
        new_md5 = self.calc_all_md5()
        try:
            with open(os.path.join(self.data_folder, 'dataset.json'), 'r') as f:
                old_md5 = json.load(f)
        except FileNotFoundError:
            return False

        return new_md5 == old_md5

    @property
    def sample_shape(self):
        """Return the sample shape info."""
        return self._sample_shape

    @property
    def target_shape(self):
        """Return the target shape info."""
        return self._target_shape

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return (f'Dogs and Cats dataset, shard number {self.rank} '
                f'out of {self.worldsize}')



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Plan module."""
from hashlib import sha384
from importlib import import_module
from logging import getLogger
from os.path import splitext
from pathlib import Path

from yaml import SafeDumper
from yaml import dump
from yaml import safe_load

from openfl.component.assigner.custom_assigner import Assigner
from openfl.interface.aggregation_functions import AggregationFunction
from openfl.interface.aggregation_functions import WeightedAverage
from openfl.interface.cli_helper import WORKSPACE
from openfl.transport import AggregatorGRPCClient
from openfl.transport import AggregatorGRPCServer
from openfl.utilities.utils import getfqdn_env

SETTINGS = 'settings'
TEMPLATE = 'template'
DEFAULTS = 'defaults'
AUTO = 'auto'


class Plan:
    """Federated Learning plan."""

    logger = getLogger(__name__)

    @staticmethod
    def load(yaml_path: Path, default: dict = None):
        """Load the plan from YAML file."""
        if default is None:
            default = {}
        if yaml_path and yaml_path.exists():
            return safe_load(yaml_path.read_text())
        return default

    @staticmethod
    def dump(yaml_path, config, freeze=False):
        """Dump the plan config to YAML file."""

        class NoAliasDumper(SafeDumper):

            def ignore_aliases(self, data):
                return True

        if freeze:
            plan = Plan()
            plan.config = config
            frozen_yaml_path = Path(
                f'{yaml_path.parent}/{yaml_path.stem}_{plan.hash[:8]}.yaml')
            if frozen_yaml_path.exists():
                Plan.logger.info(f'{yaml_path.name} is already frozen')
                return
            frozen_yaml_path.write_text(dump(config))
            frozen_yaml_path.chmod(0o400)
            Plan.logger.info(f'{yaml_path.name} frozen successfully')
        else:
            yaml_path.write_text(dump(config))

    @staticmethod
    def parse(plan_config_path: Path, cols_config_path: Path = None,
              data_config_path: Path = None, resolve=True):
        """
        Parse the Federated Learning plan.

        Args:
            plan_config_path (string): The filepath to the federated learning
                                       plan
            cols_config_path (string): The filepath to the federation
                                       collaborator list [optional]
            data_config_path (string): The filepath to the federation
                                       collaborator data configuration
                                       [optional]
        Returns:
            A federated learning plan object
        """
        try:

            plan = Plan()
            plan.config = Plan.load(plan_config_path)  # load plan configuration
            plan.name = plan_config_path.name
            plan.files = [plan_config_path]  # collect all the plan files

            # ensure 'settings' appears in each top-level section
            for section in plan.config.keys():

                if plan.config[section].get(SETTINGS) is None:
                    plan.config[section][SETTINGS] = {}

            # walk the top level keys and load 'defaults' in sorted order
            for section in sorted(plan.config.keys()):
                defaults = plan.config[section].pop(DEFAULTS, None)

                if defaults is not None:
                    defaults = WORKSPACE / 'workspace' / defaults

                    plan.files.append(defaults)

                    if resolve:
                        Plan.logger.info(
                            f'Loading DEFAULTS for section [red]{section}[/] '
                            f'from file [red]{defaults}[/].',
                            extra={'markup': True})

                    defaults = Plan.load(Path(defaults))

                    if SETTINGS in defaults:
                        # override defaults with section settings
                        defaults[SETTINGS].update(
                            plan.config[section][SETTINGS])
                        plan.config[section][SETTINGS] = defaults[SETTINGS]

                    defaults.update(plan.config[section])

                    plan.config[section] = defaults

            plan.authorized_cols = Plan.load(cols_config_path).get(
                'collaborators', []
            )

            # TODO: Does this need to be a YAML file? Probably want to use key
            #  value as the plan hash
            plan.cols_data_paths = {}
            if data_config_path is not None:
                data_config = open(data_config_path, 'r')
                for line in data_config:
                    line = line.rstrip()
                    if len(line) > 0:
                        if line[0] != '#':
                            collab, data_path = line.split(',', maxsplit=1)
                            plan.cols_data_paths[collab] = data_path

            if resolve:
                plan.resolve()

                Plan.logger.info(
                    f'Parsing Federated Learning Plan : [green]SUCCESS[/] : '
                    f'[blue]{plan_config_path}[/].',
                    extra={'markup': True})
                Plan.logger.info(dump(plan.config))

            return plan

        except Exception:
            Plan.logger.exception(f'Parsing Federated Learning Plan : '
                                  f'[red]FAILURE[/] : [blue]{plan_config_path}[/].',
                                  extra={'markup': True})
            raise

    @staticmethod
    def build(template, settings, **override):
        """
        Create an instance of a openfl Component or Federated DataLoader/TaskRunner.

        Args:
            template: Fully qualified class template path
            settings: Keyword arguments to class constructor

        Returns:
            A Python object
        """
        class_name = splitext(template)[1].strip('.')
        module_path = splitext(template)[0]

        Plan.logger.info(f'Building [red]🡆[/] Object [red]{class_name}[/] '
                         f'from [red]{module_path}[/] Module.',
                         extra={'markup': True})
        Plan.logger.debug(f'Settings [red]🡆[/] {settings}',
                          extra={'markup': True})
        Plan.logger.debug(f'Override [red]🡆[/] {override}',
                          extra={'markup': True})

        settings.update(**override)

        module = import_module(module_path)
        instance = getattr(module, class_name)(**settings)

        return instance

    @staticmethod
    def import_(template):
        """
        Import an instance of a openfl Component or Federated DataLoader/TaskRunner.

        Args:
            template: Fully qualified object path

        Returns:
            A Python object
        """
        class_name = splitext(template)[1].strip('.')
        module_path = splitext(template)[0]
        Plan.logger.info(f'Importing [red]🡆[/] Object [red]{class_name}[/] '
                         f'from [red]{module_path}[/] Module.',
                         extra={'markup': True})
        module = import_module(module_path)
        instance = getattr(module, class_name)

        return instance

    def __init__(self):
        """Initialize."""
        self.config = {}  # dictionary containing patched plan definition
        self.authorized_cols = []  # authorized collaborator list
        self.cols_data_paths = {}  # collaborator data paths dict

        self.collaborator_ = None  # collaborator object
        self.aggregator_ = None  # aggregator object
        self.assigner_ = None  # assigner object

        self.loader_ = None  # data loader object
        self.runner_ = None  # task runner object

        self.server_ = None  # gRPC server object
        self.client_ = None  # gRPC client object

        self.pipe_ = None  # compression pipeline object

        self.straggler_policy_ = None  # straggler handling policy

        self.hash_ = None
        self.name_ = None
        self.serializer_ = None

    @property
    def hash(self):  # NOQA
        """Generate hash for this instance."""
        self.hash_ = sha384(dump(self.config).encode('utf-8'))
        Plan.logger.info(f'FL-Plan hash is [blue]{self.hash_.hexdigest()}[/]',
                         extra={'markup': True})

        return self.hash_.hexdigest()

    def resolve(self):
        """Resolve the federation settings."""
        self.federation_uuid = f'{self.name}_{self.hash[:8]}'
        self.aggregator_uuid = f'aggregator_{self.federation_uuid}'

        self.rounds_to_train = self.config['aggregator'][SETTINGS][
            'rounds_to_train']

        if self.config['network'][SETTINGS]['agg_addr'] == AUTO:
            self.config['network'][SETTINGS]['agg_addr'] = getfqdn_env()

        if self.config['network'][SETTINGS]['agg_port'] == AUTO:
            self.config['network'][SETTINGS]['agg_port'] = int(
                self.hash[:8], 16
            ) % (60999 - 49152) + 49152

    def get_assigner(self, nn=True):
        """Get the plan task assigner."""
        aggregation_functions_by_task = None
        assigner_function = None
        try:
            aggregation_functions_by_task = self.restore_object('aggregation_function_obj.pkl')
            assigner_function = self.restore_object('task_assigner_obj.pkl')
        except Exception as exc:
            self.logger.error(f'Failed to load aggregation and assigner functions: {exc}')
            self.logger.info('Using Task Runner API workflow')
        if assigner_function and nn:
            self.assigner_ = Assigner(
                assigner_function=assigner_function,
                aggregation_functions_by_task=aggregation_functions_by_task,
                authorized_cols=self.authorized_cols,
                rounds_to_train=self.rounds_to_train,
            )
        else:
            # Backward compatibility
            defaults = self.config.get(
                'assigner',
                {
                    TEMPLATE: 'openfl.component.Assigner',
                    SETTINGS: {}
                }
            )

            defaults[SETTINGS]['authorized_cols'] = self.authorized_cols
            defaults[SETTINGS]['rounds_to_train'] = self.rounds_to_train
            defaults[SETTINGS]['tasks'] = self.get_tasks()

            if self.assigner_ is None:
                self.assigner_ = Plan.build(**defaults)

        return self.assigner_

    def get_tasks(self):
        """Get federation tasks."""
        tasks = self.config.get('tasks', {})
        tasks.pop(DEFAULTS, None)
        tasks.pop(SETTINGS, None)
        for task in tasks:
            aggregation_type = tasks[task].get('aggregation_type')
            if aggregation_type is None:
                aggregation_type = WeightedAverage()
            elif isinstance(aggregation_type, dict):
                if SETTINGS not in aggregation_type:
                    aggregation_type[SETTINGS] = {}
                aggregation_type = Plan.build(**aggregation_type)
                if not isinstance(aggregation_type, AggregationFunction):
                    raise NotImplementedError(f'''{task} task aggregation type does not implement an interface:
    openfl.interface.aggregation_functions.AggregationFunction
    ''')
            tasks[task]['aggregation_type'] = aggregation_type
        return tasks

    def get_aggregator(self, tensor_dict=None):
        """Get federation aggregator."""
        defaults = self.config.get('aggregator',
                                   {
                                       TEMPLATE: 'openfl.component.Aggregator',
                                       SETTINGS: {}
                                   })

        defaults[SETTINGS]['aggregator_uuid'] = self.aggregator_uuid
        defaults[SETTINGS]['federation_uuid'] = self.federation_uuid
        defaults[SETTINGS]['authorized_cols'] = self.authorized_cols
        nn = defaults[SETTINGS]['nn'] if 'nn' in defaults[SETTINGS] else True
        defaults[SETTINGS]['assigner'] = self.get_assigner(nn)
        defaults[SETTINGS]['compression_pipeline'] = self.get_tensor_pipe()
        defaults[SETTINGS]['straggler_handling_policy'] = self.get_straggler_handling_policy()
        log_metric_callback = defaults[SETTINGS].get('log_metric_callback')

        if log_metric_callback:
            if isinstance(log_metric_callback, dict):
                log_metric_callback = Plan.import_(**log_metric_callback)
            elif not callable(log_metric_callback):
                raise TypeError(f'log_metric_callback should be callable object '
                                f'or be import from code part, get {log_metric_callback}')

        defaults[SETTINGS]['log_metric_callback'] = log_metric_callback
        if self.aggregator_ is None:
            self.aggregator_ = Plan.build(**defaults, initial_tensor_dict=tensor_dict)

        return self.aggregator_

    def get_tensor_pipe(self):
        """Get data tensor pipeline."""
        defaults = self.config.get(
            'compression_pipeline',
            {
                TEMPLATE: 'openfl.pipelines.NoCompressionPipeline',
                SETTINGS: {}
            }
        )

        if self.pipe_ is None:
            self.pipe_ = Plan.build(**defaults)

        return self.pipe_

    def get_straggler_handling_policy(self):
        """Get straggler handling policy."""
        defaults = self.config.get(
            'straggler_handling_policy',
            {
                TEMPLATE: 'openfl.component.straggler_handling_functions.CutoffTimeBasedStragglerHandling',
                SETTINGS: {}
            }
        )

        if self.straggler_policy_ is None:
            self.straggler_policy_ = Plan.build(**defaults)

        return self.straggler_policy_

    # legacy api (TaskRunner subclassing)
    def get_data_loader(self, collaborator_name):
        """Get data loader."""
        defaults = self.config.get('data_loader',
                                   {
                                       TEMPLATE: 'openfl.federation.DataLoader',
                                       SETTINGS: {}
                                   })

        defaults[SETTINGS]['data_path'] = self.cols_data_paths[
            collaborator_name
        ]

        if self.loader_ is None:
            self.loader_ = Plan.build(**defaults)

        return self.loader_

    # Python interactive api
    def initialize_data_loader(self, data_loader, shard_descriptor):
        """Get data loader."""
        data_loader.shard_descriptor = shard_descriptor
        return data_loader

    # legacy api (TaskRunner subclassing)
    def get_task_runner(self, data_loader):
        """Get task runner."""
        defaults = self.config.get('task_runner',
                                   {
                                       TEMPLATE: 'openfl.federation.TaskRunner',
                                       SETTINGS: {}
                                   })

        defaults[SETTINGS]['data_loader'] = data_loader

        if self.runner_ is None:
            self.runner_ = Plan.build(**defaults)

        return self.runner_

    # Python interactive api
    # TODO: remove nn from the parameters
    def get_core_task_runner(self, data_loader=None,
                             model_provider=None,
                             task_keeper=None,
                             nn=True):
        """Get task runner."""
        defaults = self.config.get(
            'task_runner',
            {
                TEMPLATE: 'openfl.federated.task.task_runner.CoreTaskRunner',
                SETTINGS: {}
            })
        defaults[SETTINGS]['nn'] = nn

        # We are importing a CoreTaskRunner instance!!!
        if self.runner_ is None:
            self.runner_ = Plan.build(**defaults)

        self.runner_.set_data_loader(data_loader)

        self.runner_.set_model_provider(model_provider)
        self.runner_.set_task_provider(task_keeper)

        framework_adapter = Plan.build(
            self.config['task_runner']['required_plugin_components']['framework_adapters'], {})

        # This step initializes tensorkeys
        # Which have no sens if task provider is not set up
        self.runner_.set_framework_adapter(framework_adapter)

        return self.runner_

    def get_collaborator(self, collaborator_name, root_certificate=None, private_key=None,
                         certificate=None, task_runner=None, client=None, shard_descriptor=None):
        """Get collaborator."""
        defaults = self.config.get(
            'collaborator',
            {
                TEMPLATE: 'openfl.component.Collaborator',
                SETTINGS: {}
            }
        )

        defaults[SETTINGS]['collaborator_name'] = collaborator_name
        defaults[SETTINGS]['aggregator_uuid'] = self.aggregator_uuid
        defaults[SETTINGS]['federation_uuid'] = self.federation_uuid
        defaults[SETTINGS]['nn'] = defaults[SETTINGS]['nn'] if 'nn' in defaults[SETTINGS] else True

        if task_runner is not None:
            defaults[SETTINGS]['task_runner'] = task_runner
        else:
            # Here we support new interactive api as well as old task_runner subclassing interface
            # If Task Runner class is placed inside openfl `task-runner` subpackage it is
            # a part of the New API and it is a part of OpenFL kernel.
            # If Task Runner is placed elsewhere, somewhere in user workspace, than it is
            # a part of the old interface and we follow legacy initialization procedure.
            # TODO: this switch for AdaBoost should be implemented in a better way
            if 'openfl.federated.task.task_runner' in self.config['task_runner']['template'] \
                    or 'openfl.federated.task.runner_generic.GenericTaskRunner' in self.config['task_runner'][
                'template']:
                # Interactive API
                model_provider, task_keeper, data_loader = self.deserialize_interface_objects()
                data_loader = self.initialize_data_loader(data_loader, shard_descriptor)
                defaults[SETTINGS]['task_runner'] = self.get_core_task_runner(
                    data_loader=data_loader,
                    model_provider=model_provider,
                    task_keeper=task_keeper,
                    nn=defaults[SETTINGS]['nn'])
            else:
                # TaskRunner subclassing API
                data_loader = self.get_data_loader(collaborator_name)
                defaults[SETTINGS]['task_runner'] = self.get_task_runner(data_loader)

        defaults[SETTINGS]['compression_pipeline'] = self.get_tensor_pipe()
        defaults[SETTINGS]['task_config'] = self.config.get('tasks', {})
        if client is not None:
            defaults[SETTINGS]['client'] = client
        else:
            defaults[SETTINGS]['client'] = self.get_client(
                collaborator_name,
                self.aggregator_uuid,
                self.federation_uuid,
                root_certificate,
                private_key,
                certificate
            )

        if self.collaborator_ is None:
            self.collaborator_ = Plan.build(**defaults)

        return self.collaborator_

    def get_client(self, collaborator_name, aggregator_uuid, federation_uuid,
                   root_certificate=None, private_key=None, certificate=None):
        """Get gRPC client for the specified collaborator."""
        common_name = collaborator_name
        if not root_certificate or not private_key or not certificate:
            root_certificate = 'cert/cert_chain.crt'
            certificate = f'cert/client/col_{common_name}.crt'
            private_key = f'cert/client/col_{common_name}.key'

        client_args = self.config['network'][SETTINGS]

        # patch certificates

        client_args['root_certificate'] = root_certificate
        client_args['certificate'] = certificate
        client_args['private_key'] = private_key

        client_args['aggregator_uuid'] = aggregator_uuid
        client_args['federation_uuid'] = federation_uuid

        if self.client_ is None:
            self.client_ = AggregatorGRPCClient(**client_args)

        return self.client_

    def get_server(self, root_certificate=None, private_key=None, certificate=None, **kwargs):
        """Get gRPC server of the aggregator instance."""
        common_name = self.config['network'][SETTINGS]['agg_addr'].lower()

        if not root_certificate or not private_key or not certificate:
            root_certificate = 'cert/cert_chain.crt'
            certificate = f'cert/server/agg_{common_name}.crt'
            private_key = f'cert/server/agg_{common_name}.key'

        server_args = self.config['network'][SETTINGS]

        # patch certificates

        server_args.update(kwargs)
        server_args['root_certificate'] = root_certificate
        server_args['certificate'] = certificate
        server_args['private_key'] = private_key

        server_args['aggregator'] = self.get_aggregator()

        if self.server_ is None:
            self.server_ = AggregatorGRPCServer(**server_args)

        return self.server_

    def interactive_api_get_server(self, *, tensor_dict, root_certificate, certificate,
                                   private_key, tls):
        """Get gRPC server of the aggregator instance."""
        server_args = self.config['network'][SETTINGS]

        # patch certificates
        server_args['root_certificate'] = root_certificate
        server_args['certificate'] = certificate
        server_args['private_key'] = private_key
        server_args['tls'] = tls

        server_args['aggregator'] = self.get_aggregator(tensor_dict)

        if self.server_ is None:
            self.server_ = AggregatorGRPCServer(**server_args)

        return self.server_

    def deserialize_interface_objects(self):
        """Deserialize objects for TaskRunner."""
        api_layer = self.config['api_layer']
        filenames = [
            'model_interface_file',
            'tasks_interface_file',
            'dataloader_interface_file'
        ]
        return (self.restore_object(api_layer['settings'][filename]) for filename in filenames)

    def get_serializer_plugin(self, **kwargs):
        """Get serializer plugin.

        This plugin is used for serialization of interfaces in new interactive API
        """
        if self.serializer_ is None:
            if 'api_layer' not in self.config:  # legacy API
                return None
            required_plugin_components = self.config['api_layer']['required_plugin_components']
            serializer_plugin = required_plugin_components['serializer_plugin']
            self.serializer_ = Plan.build(serializer_plugin, kwargs)
        return self.serializer_

    def restore_object(self, filename):
        """Deserialize an object."""
        serializer_plugin = self.get_serializer_plugin()
        if serializer_plugin is None:
            return None
        obj = serializer_plugin.restore_object(filename)
        return obj



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Kvasir shard descriptor."""

import os
from pathlib import Path

import numpy as np
from PIL import Image

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor
from openfl.utilities import validate_file_hash


class KvasirShardDataset(ShardDataset):
    """Kvasir Shard dataset class."""

    def __init__(self, dataset_dir: Path, rank=1, worldsize=1, enforce_image_hw=None):
        """Initialize KvasirShardDataset."""
        self.rank = rank
        self.worldsize = worldsize
        self.dataset_dir = dataset_dir
        self.enforce_image_hw = enforce_image_hw
        self.images_path = self.dataset_dir / 'segmented-images' / 'images'
        self.masks_path = self.dataset_dir / 'segmented-images' / 'masks'

        self.images_names = [
            img_name
            for img_name in sorted(os.listdir(self.images_path))
            if len(img_name) > 3 and img_name[-3:] == 'jpg'
        ]
        # Sharding
        self.images_names = self.images_names[self.rank - 1::self.worldsize]

    def __getitem__(self, index):
        """Return a item by the index."""
        name = self.images_names[index]
        # Reading data
        img = Image.open(self.images_path / name)
        mask = Image.open(self.masks_path / name)
        if self.enforce_image_hw is not None:
            # If we need to resize data
            # PIL accepts (w,h) tuple, not (h,w)
            img = img.resize(self.enforce_image_hw[::-1])
            mask = mask.resize(self.enforce_image_hw[::-1])
        img = np.asarray(img)
        mask = np.asarray(mask)
        assert img.shape[2] == 3

        return img, mask[:, :, 0].astype(np.uint8)

    def __len__(self):
        """Return the len of the dataset."""
        return len(self.images_names)


class KvasirShardDescriptor(ShardDescriptor):
    """Shard descriptor class."""

    def __init__(self, data_folder: str = 'kvasir_data',
                 rank_worldsize: str = '1,1',
                 enforce_image_hw: str = None) -> None:
        """Initialize KvasirShardDescriptor."""
        super().__init__()
        # Settings for sharding the dataset
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))

        self.data_folder = Path.cwd() / data_folder
        self.download_data(self.data_folder)

        # Settings for resizing data
        self.enforce_image_hw = None
        if enforce_image_hw is not None:
            self.enforce_image_hw = tuple(int(size) for size in enforce_image_hw.split(','))

        # Calculating data and target shapes
        ds = self.get_dataset()
        sample, target = ds[0]
        self._sample_shape = [str(dim) for dim in sample.shape]
        self._target_shape = [str(dim) for dim in target.shape]

    def get_dataset(self, dataset_type='train'):
        """Return a shard dataset by type."""
        return KvasirShardDataset(
            dataset_dir=self.data_folder,
            rank=self.rank,
            worldsize=self.worldsize,
            enforce_image_hw=self.enforce_image_hw
        )

    @staticmethod
    def download_data(data_folder):
        """Download data."""
        zip_file_path = data_folder / 'kvasir.zip'
        os.makedirs(data_folder, exist_ok=True)
        os.system(
            'wget -nc'
            " 'https://datasets.simula.no/downloads/"
            "hyper-kvasir/hyper-kvasir-segmented-images.zip'"
            f' -O {zip_file_path.relative_to(Path.cwd())}'
        )
        zip_sha384 = ('66cd659d0e8afd8c83408174'
                      '1ade2b75dada8d4648b816f2533c8748b1658efa3d49e205415d4116faade2c5810e241e')
        validate_file_hash(zip_file_path, zip_sha384)
        os.system(f'unzip -n {zip_file_path.relative_to(Path.cwd())}'
                  f' -d {data_folder.relative_to(Path.cwd())}')

    @property
    def sample_shape(self):
        """Return the sample shape info."""
        return self._sample_shape

    @property
    def target_shape(self):
        """Return the target shape info."""
        return self._target_shape

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return (f'Kvasir dataset, shard number {self.rank} '
                f'out of {self.worldsize}')


if __name__ == '__main__':
    from openfl.interface.cli import setup_logging

    setup_logging()

    data_folder = 'data'
    rank_worldsize = '1,100'
    enforce_image_hw = '529,622'

    kvasir_sd = KvasirShardDescriptor(
        data_folder,
        rank_worldsize=rank_worldsize,
        enforce_image_hw=enforce_image_hw)

    print(kvasir_sd.dataset_description)
    print(kvasir_sd.sample_shape, kvasir_sd.target_shape)

    from openfl.component.envoy.envoy import Envoy

    shard_name = 'one'
    director_host = 'localhost'
    director_port = 50051

    keeper = Envoy(
        shard_name=shard_name,
        director_host=director_host,
        director_port=director_port,
        shard_descriptor=kvasir_sd,
        tls=True,
        root_certificate='./cert/root_ca.crt',
        private_key='./cert/one.key',
        certificate='./cert/one.crt',
    )

    keeper.start()



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Experiment module."""

import asyncio
import logging
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Iterable
from typing import List
from typing import Union

from openfl.federated import Plan
from openfl.transport import AggregatorGRPCServer
from openfl.utilities.workspace import ExperimentWorkspace

logger = logging.getLogger(__name__)


class Status:
    """Experiment's statuses."""

    PENDING = 'pending'
    FINISHED = 'finished'
    IN_PROGRESS = 'in_progress'
    FAILED = 'failed'


class Experiment:
    """Experiment class."""

    def __init__(
            self, *,
            name: str,
            archive_path: Union[Path, str],
            collaborators: List[str],
            sender: str,
            init_tensor_dict: dict,
            plan_path: Union[Path, str] = 'plan/plan.yaml',
            users: Iterable[str] = None,
    ) -> None:
        """Initialize an experiment object."""
        self.name = name
        if isinstance(archive_path, str):
            archive_path = Path(archive_path)
        self.archive_path = archive_path
        self.collaborators = collaborators
        self.sender = sender
        self.init_tensor_dict = init_tensor_dict
        if isinstance(plan_path, str):
            plan_path = Path(plan_path)
        self.plan_path = plan_path
        self.users = set() if users is None else set(users)
        self.status = Status.PENDING
        self.aggregator = None

    async def start(
            self, *,
            tls: bool = True,
            root_certificate: Union[Path, str] = None,
            private_key: Union[Path, str] = None,
            certificate: Union[Path, str] = None,
    ):
        """Run experiment."""
        self.status = Status.IN_PROGRESS
        try:
            logger.info(f'New experiment {self.name} for '
                        f'collaborators {self.collaborators}')

            with ExperimentWorkspace(self.name, self.archive_path):
                aggregator_grpc_server = self._create_aggregator_grpc_server(
                    tls=tls,
                    root_certificate=root_certificate,
                    private_key=private_key,
                    certificate=certificate,
                )
                self.aggregator = aggregator_grpc_server.aggregator
                await self._run_aggregator_grpc_server(
                    aggregator_grpc_server=aggregator_grpc_server,
                )
            self.status = Status.FINISHED
            logger.info(f'Experiment "{self.name}" was finished successfully.')
        except Exception as e:
            self.status = Status.FAILED
            logger.exception(f'Experiment "{self.name}" was failed with error: {e}.')

    def _create_aggregator_grpc_server(
            self, *,
            tls: bool = True,
            root_certificate: Union[Path, str] = None,
            private_key: Union[Path, str] = None,
            certificate: Union[Path, str] = None,
    ) -> AggregatorGRPCServer:
        plan = Plan.parse(plan_config_path=Path(self.plan_path))
        plan.authorized_cols = list(self.collaborators)

        logger.info('🧿 Starting the Aggregator Service.')
        aggregator_grpc_server = plan.interactive_api_get_server(
            tensor_dict=self.init_tensor_dict,
            root_certificate=root_certificate,
            certificate=certificate,
            private_key=private_key,
            tls=tls,
        )
        return aggregator_grpc_server

    @staticmethod
    async def _run_aggregator_grpc_server(aggregator_grpc_server: AggregatorGRPCServer) -> None:
        """Run aggregator."""
        logger.info('🧿 Starting the Aggregator Service.')
        grpc_server = aggregator_grpc_server.get_server()
        grpc_server.start()
        logger.info('Starting Aggregator gRPC Server')

        try:
            while not aggregator_grpc_server.aggregator.all_quit_jobs_sent():
                # Awaiting quit job sent to collaborators
                await asyncio.sleep(10)
        except KeyboardInterrupt:
            pass
        finally:
            grpc_server.stop(0)
            # Temporary solution to free RAM used by TensorDB
            aggregator_grpc_server.aggregator.tensor_db.clean_up(0)


class ExperimentsRegistry:
    """ExperimentsList class."""

    def __init__(self) -> None:
        """Initialize an experiments list object."""
        self.__active_experiment_name = None
        self.__pending_experiments = []
        self.__archived_experiments = []
        self.__dict = {}

    @property
    def active_experiment(self) -> Union[Experiment, None]:
        """Get active experiment."""
        if self.__active_experiment_name is None:
            return None
        return self.__dict[self.__active_experiment_name]

    @property
    def pending_experiments(self) -> List[str]:
        """Get queue of not started experiments."""
        return self.__pending_experiments

    def add(self, experiment: Experiment) -> None:
        """Add experiment to queue of not started experiments."""
        self.__dict[experiment.name] = experiment
        self.__pending_experiments.append(experiment.name)

    def remove(self, name: str) -> None:
        """Remove experiment from everywhere."""
        if self.__active_experiment_name == name:
            self.__active_experiment_name = None
        if name in self.__pending_experiments:
            self.__pending_experiments.remove(name)
        if name in self.__archived_experiments:
            self.__archived_experiments.remove(name)
        if name in self.__dict:
            del self.__dict[name]

    def __getitem__(self, key: str) -> Experiment:
        """Get experiment by name."""
        return self.__dict[key]

    def get(self, key: str, default=None) -> Experiment:
        """Get experiment by name."""
        return self.__dict.get(key, default)

    def get_user_experiments(self, user: str) -> List[Experiment]:
        """Get list of experiments for specific user."""
        return [
            exp
            for exp in self.__dict.values()
            if user in exp.users
        ]

    def __contains__(self, key: str) -> bool:
        """Check if experiment exists."""
        return key in self.__dict

    def finish_active(self) -> None:
        """Finish active experiment."""
        self.__archived_experiments.insert(0, self.__active_experiment_name)
        self.__active_experiment_name = None

    @asynccontextmanager
    async def get_next_experiment(self):
        """Context manager.

        On enter get experiment from pending_experiments.
        On exit put finished experiment to archive_experiments.
        """
        while True:
            if self.active_experiment is None and self.pending_experiments:
                break
            await asyncio.sleep(10)

        try:
            self.__active_experiment_name = self.pending_experiments.pop(0)
            yield self.active_experiment
        finally:
            self.finish_active()



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Adam adaptive aggregation module."""

from typing import Dict
from typing import Optional
from typing import Tuple

import numpy as np

from openfl.utilities.optimizers.numpy import NumPyAdam
from .core import AdaptiveAggregation
from .core import AggregationFunction
from .weighted_average import WeightedAverage


DEFAULT_AGG_FUNC = WeightedAverage()


class AdamAdaptiveAggregation(AdaptiveAggregation):
    """Adam adaptive Federated Aggregation funtcion."""

    def __init__(
        self,
        *,
        agg_func: AggregationFunction = DEFAULT_AGG_FUNC,
        params: Optional[Dict[str, np.ndarray]] = None,
        model_interface=None,
        learning_rate: float = 0.01,
        betas: Tuple[float, float] = (0.9, 0.999),
        initial_accumulator_value: float = 0.0,
        epsilon: float = 1e-8,
    ) -> None:
        """Initialize.

        Args:
            agg_func: Aggregate function for aggregating
                parameters that are not inside the optimizer (default: WeightedAverage()).
            params: Parameters to be stored for optimization.
            model_interface: Model interface instance to provide parameters.
            learning_rate: Tuning parameter that determines
                the step size at each iteration.
            betas: Coefficients used for computing running
                averages of gradient and its square.
            initial_accumulator_value: Initial value for gradients
                and squared gradients.
            epsilon: Value for computational stability.
        """
        opt = NumPyAdam(params=params,
                        model_interface=model_interface,
                        learning_rate=learning_rate,
                        betas=betas,
                        initial_accumulator_value=initial_accumulator_value,
                        epsilon=epsilon)
        super().__init__(opt, agg_func)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Aggregation function interface module."""
from abc import abstractmethod
from typing import Iterator
from typing import List
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from openfl.utilities import LocalTensor
from openfl.interface.aggregation_functions import AggregationFunction


class PrivilegedAggregationFunction(AggregationFunction):
    """Privileged Aggregation Function interface provides write access to TensorDB Dataframe.

    """

    def __init__(
        self
    ) -> None:
        """Initialize with TensorDB write access"""
        super().__init__()
        self._privileged = True

    @abstractmethod
    def call(self,
             local_tensors: List[LocalTensor],
             tensor_db: pd.DataFrame,
             tensor_name: str,
             fl_round: int,
             tags: Tuple[str]) -> np.ndarray:
        """Aggregate tensors.

        Args:
            local_tensors(list[openfl.utilities.LocalTensor]): List of local tensors to aggregate.
            tensor_db: Raw TensorDB dataframe (for write access). Columns:
                - 'tensor_name': name of the tensor.
                    Examples for `torch.nn.Module`s: 'conv1.weight', 'fc2.bias'.
                - 'round': 0-based number of round corresponding to this tensor.
                - 'tags': tuple of tensor tags. Tags that can appear:
                    - 'model' indicates that the tensor is a model parameter.
                    - 'trained' indicates that tensor is a part of a training result.
                        These tensors are passed to the aggregator node after local learning.
                    - 'aggregated' indicates that tensor is a result of aggregation.
                        These tensors are sent to collaborators for the next round.
                    - 'delta' indicates that value is a difference between rounds
                        for a specific tensor.
                    also one of the tags is a collaborator name
                    if it corresponds to a result of a local task.

                - 'nparray': value of the tensor.
            tensor_name: name of the tensor
            fl_round: round number
            tags: tuple of tags for this tensor
        Returns:
            np.ndarray: aggregated tensor
        """
        raise NotImplementedError




# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""openfl.utilities.data_splitters.data_splitter module."""
from abc import ABC
from abc import abstractmethod
from typing import Iterable
from typing import List
from typing import TypeVar

T = TypeVar('T')


class DataSplitter(ABC):
    """Base class for data splitting."""

    @abstractmethod
    def split(self, data: Iterable[T], num_collaborators: int) -> List[Iterable[T]]:
        """Split the data."""
        raise NotImplementedError



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import zipfile
from os import listdir
from pathlib import Path

import numpy as np
import PIL
from skimage import io
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from torchvision import transforms as tsf
from torchvision.datasets.utils import download_url
from tqdm import tqdm

from openfl.federated import PyTorchDataLoader
from openfl.utilities import validate_file_hash


def read_data(image_path, mask_path):
    """Read image and mask from disk.

    Args:
        image_path: Path to image
        mask_path:  Path to mask

    Returns:
        Numpy image and mask

    """
    img = io.imread(image_path)
    assert (img.shape[2] == 3)
    mask = io.imread(mask_path)
    return (img, mask[:, :, 0].astype(np.uint8))


class KvasirDataset(Dataset):
    """Kvasir dataset. Splits data by shards for each collaborator."""

    def __init__(self, is_validation, shard_num, collaborator_count, **kwargs):
        """Initialize dataset.

        Args:
            is_validation: Validation dataset or not
            shard_num: Number of collaborator for which the data is splited
            collaborator_count: Total number of collaborators

        """
        self.images_path = './data/segmented-images/images/'
        self.masks_path = './data/segmented-images/masks/'
        self.images_names = [
            img_name
            for img_name in sorted(listdir(self.images_path))
            if len(img_name) > 3 and img_name[-3:] == 'jpg'
        ]

        self.images_names = self.images_names[shard_num:: collaborator_count]
        self.is_validation = is_validation
        assert (len(self.images_names) > 8)
        validation_size = len(self.images_names) // 8

        if is_validation:
            self.images_names = self.images_names[-validation_size:]
        else:
            self.images_names = self.images_names[: -validation_size]

        self.img_trans = tsf.Compose([
            tsf.ToPILImage(),
            tsf.Resize((332, 332)),
            tsf.ToTensor(),
            tsf.Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])])
        self.mask_trans = tsf.Compose([
            tsf.ToPILImage(),
            tsf.Resize((332, 332), interpolation=PIL.Image.NEAREST),
            tsf.ToTensor()])

    def __getitem__(self, index):
        """Get items by slice index."""
        name = self.images_names[index]
        img, mask = read_data(self.images_path + name, self.masks_path + name)
        img = self.img_trans(img)
        mask = self.mask_trans(mask)
        return img, mask

    def __len__(self):
        """Length of spltted data."""
        return len(self.images_names)


def load_kvasir_dataset():
    """Load and unzip kvasir dataset."""
    zip_sha384 = ('66cd659d0e8afd8c83408174'
                  '1ade2b75dada8d4648b816f2533c8748b1658efa3d49e205415d4116faade2c5810e241e')
    data_url = ('https://datasets.simula.no/downloads/'
                'hyper-kvasir/hyper-kvasir-segmented-images.zip')
    filename = 'kvasir.zip'
    data_folder_path = Path.cwd().absolute() / 'data'
    kvasir_archive_path = data_folder_path / filename
    if not kvasir_archive_path.is_file():
        download_url(data_url, data_folder_path, filename=filename)
        validate_file_hash(kvasir_archive_path, zip_sha384)
        with zipfile.ZipFile(kvasir_archive_path, 'r') as zip_ref:
            for member in tqdm(iterable=zip_ref.infolist(), desc='Unzipping dataset'):
                zip_ref.extract(member, './data')


class PyTorchKvasirDataLoader(PyTorchDataLoader):
    """PyTorch data loader for Kvasir dataset."""

    def __init__(self, data_path, batch_size, **kwargs):
        """Instantiate the data object.

        Args:
            data_path: The file path to the data
            batch_size: The batch size of the data loader
            **kwargs: Additional arguments, passed to super
             init and load_mnist_shard
        """
        super().__init__(batch_size, **kwargs)

        load_kvasir_dataset()
        self.valid_dataset = KvasirDataset(True, shard_num=int(data_path), **kwargs)
        self.train_dataset = KvasirDataset(False, shard_num=int(data_path), **kwargs)
        self.train_loader = self.get_train_loader()
        self.val_loader = self.get_valid_loader()
        self.batch_size = batch_size

    def get_valid_loader(self, num_batches=None):
        """Return validation dataloader."""
        return DataLoader(self.valid_dataset, batch_size=self.batch_size)

    def get_train_loader(self, num_batches=None):
        """Return train dataloader."""
        return DataLoader(self.train_dataset, batch_size=self.batch_size, shuffle=True)

    def get_train_data_size(self):
        """Return size of train dataset."""
        return len(self.train_dataset)

    def get_valid_data_size(self):
        """Return size of validation dataset."""
        return len(self.valid_dataset)

    def get_feature_shape(self):
        """Return data shape."""
        return self.valid_dataset[0][0].shape



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Cutoff time based Straggler Handling function."""
import numpy as np
import time

from openfl.component.straggler_handling_functions import StragglerHandlingFunction


class CutoffTimeBasedStragglerHandling(StragglerHandlingFunction):
    def __init__(
        self,
        round_start_time=None,
        straggler_cutoff_time=np.inf,
        minimum_reporting=1,
        **kwargs):
        self.round_start_time = round_start_time
        self.straggler_cutoff_time = straggler_cutoff_time
        self.minimum_reporting = minimum_reporting

    def straggler_time_expired(self):
        return self.round_start_time is not None and (
            (time.time() - self.round_start_time) > self.straggler_cutoff_time)

    def minimum_collaborators_reported(self, num_collaborators_done):
        return num_collaborators_done >= self.minimum_reporting

    def straggler_cutoff_check(self, num_collaborators_done, all_collaborators=None):
        cutoff = self.straggler_time_expired() and self.minimum_collaborators_reported(
            num_collaborators_done)
        return cutoff



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Market shard descriptor."""

import logging
import re
import zipfile
from pathlib import Path
from typing import List

import gdown
from PIL import Image

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor

logger = logging.getLogger(__name__)


class MarketShardDataset(ShardDataset):
    """Market shard dataset."""

    def __init__(self, dataset_dir: Path, dataset_type: str, rank=1, worldsize=1):
        """Initialize MarketShardDataset."""
        self.dataset_dir = dataset_dir
        self.dataset_type = dataset_type
        self.rank = rank
        self.worldsize = worldsize

        self.imgs_path = list(dataset_dir.glob('*.jpg'))[self.rank - 1::self.worldsize]
        self.pattern = re.compile(r'([-\d]+)_c(\d)')

    def __len__(self):
        """Length of shard."""
        return len(self.imgs_path)

    def __getitem__(self, index: int):
        """Return an item by the index."""
        img_path = self.imgs_path[index]
        pid, camid = map(int, self.pattern.search(img_path.name).groups())

        img = Image.open(img_path)
        return img, (pid, camid)


class MarketShardDescriptor(ShardDescriptor):
    """
    Market1501 Shard descriptor class.

    Reference:
    Zheng et al. Scalable Person Re-identification: A Benchmark. ICCV 2015.
    URL: http://www.liangzheng.org/Project/project_reid.html

    Dataset statistics:
        identities: 1501 (+1 for background)
        images: 12936 (train) + 3368 (query) + 15913 (gallery)
    """

    def __init__(self, data_folder_name: str = 'Market-1501-v15.09.15',
                 rank_worldsize: str = '1,1') -> None:
        """Initialize MarketShardDescriptor."""
        super().__init__()

        # Settings for sharding the dataset
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))

        self.data_folder_name = data_folder_name
        self.dataset_dir = Path.cwd() / data_folder_name
        self.download()

        self.path_by_type = {
            'train': self.dataset_dir / 'bounding_box_train',
            'query': self.dataset_dir / 'query',
            'gallery': self.dataset_dir / 'bounding_box_test'
        }
        self._check_before_run()

    def get_shard_dataset_types(self) -> List[str]:
        """Get available shard dataset types."""
        return list(self.path_by_type)

    def get_dataset(self, dataset_type='train'):
        """Return a dataset by type."""
        if dataset_type not in self.path_by_type:
            raise Exception(f'Wrong dataset type: {dataset_type}.'
                            f'Choose from the list: {", ".join(self.path_by_type)}')
        return MarketShardDataset(
            dataset_dir=self.path_by_type[dataset_type],
            dataset_type=dataset_type,
            rank=self.rank,
            worldsize=self.worldsize
        )

    @property
    def sample_shape(self):
        """Return the sample shape info."""
        return ['64', '128', '3']

    @property
    def target_shape(self):
        """Return the target shape info."""
        return ['2']

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return (f'Market dataset, shard number {self.rank} '
                f'out of {self.worldsize}')

    def download(self):
        """Download Market1501 dataset."""
        logger.info('Download Market1501 dataset.')
        if self.dataset_dir.exists():
            return None

        logger.info('Try to download.')
        output = f'{self.data_folder_name}.zip'

        if not Path(output).exists():
            url = 'https://drive.google.com/u/1/uc?id=0B8-rUzbwVRk0c054eEozWG9COHM'
            gdown.download(url, output, quiet=False)
        logger.info(f'{output} is downloaded.')

        with zipfile.ZipFile(output, 'r') as zip_ref:
            zip_ref.extractall(Path.cwd())

        Path(output).unlink()  # remove zip

    def _check_before_run(self):
        """Check if all files are available before going deeper."""
        if not self.dataset_dir.exists():
            raise RuntimeError(f'{self.dataset_dir} does not exist')
        for dataset_path in self.path_by_type.values():
            if not dataset_path.exists():
                raise RuntimeError(f'{dataset_path} does not exist')



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""openfl.component package."""

from .aggregator import Aggregator
from .assigner import Assigner
from .assigner import RandomGroupedAssigner
from .assigner import StaticGroupedAssigner
from .collaborator import Collaborator
from .straggler_handling_functions import StragglerHandlingFunction
from .straggler_handling_functions import CutoffTimeBasedStragglerHandling
from .straggler_handling_functions import PercentageBasedStragglerHandling

__all__ = [
    'Assigner',
    'RandomGroupedAssigner',
    'StaticGroupedAssigner',
    'Aggregator',
    'Collaborator',
    'StragglerHandlingFunction',
    'CutoffTimeBasedStragglerHandling',
    'PercentageBasedStragglerHandling'
]



# Copyright (C) 2020-2021 Intel Corporation
# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""openfl.pipelines module."""

import pkgutil

if pkgutil.find_loader('torch'):
    from .eden_pipeline import EdenPipeline  # NOQA

from .generic_pipeline import GenericPipeline
from .kc_pipeline import KCPipeline
from .no_compression_pipeline import NoCompressionPipeline
from .random_shift_pipeline import RandomShiftPipeline
from .skc_pipeline import SKCPipeline
from .stc_pipeline import STCPipeline
from .tensor_codec import TensorCodec

__all__ = [
    'NoCompressionPipeline',
    'RandomShiftPipeline',
    'STCPipeline',
    'SKCPipeline',
    'KCPipeline',
    'EdenPipeline',
    'GenericPipeline',
    'TensorCodec',
]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""FederatedModel module."""

import inspect

from .runner import TaskRunner


class FederatedModel(TaskRunner):
    """
    A wrapper that adapts to Tensorflow and Pytorch models to a federated context.

    Args:
        model : tensorflow/keras (function) , pytorch (class)
            For keras/tensorflow model, expects a function that returns the
            model definition
            For pytorch models, expects a class (not an instance) containing
            the model definition and forward function
        optimizer : lambda function (only required for pytorch)
            The optimizer should be definied within a lambda function. This
            allows the optimizer to be attached to the federated models spawned
            for each collaborator.
        loss_fn : pytorch loss_fun (only required for pytorch)
    """

    def __init__(self, build_model,  nn=False, optimizer=None, loss_fn=None, **kwargs):
        """Initialize.

        Args:
            model:    build_model function
            **kwargs: Additional parameters to pass to the function

        """
        super().__init__(**kwargs)

        if nn:
            self.build_model = build_model
            self.lambda_opt = None
            # TODO pass params to model
            if inspect.isclass(build_model):
                self.model = build_model()
                from .runner_pt import PyTorchTaskRunner
                if optimizer is not None:
                    self.optimizer = optimizer(self.model.parameters())
                self.runner = PyTorchTaskRunner(**kwargs)
                if hasattr(self.model, 'forward'):
                    self.runner.forward = self.model.forward
            else:
                self.model = self.build_model(
                    self.feature_shape, self.data_loader.num_classes)
                from .runner_keras import KerasTaskRunner
                self.runner = KerasTaskRunner(**kwargs)
                self.optimizer = self.model.optimizer
            self.lambda_opt = optimizer
            if hasattr(self.model, 'validate'):
                self.runner.validate = lambda *args, **kwargs: build_model.validate(
                    self.runner, *args, **kwargs)
            if hasattr(self.model, 'train_epoch'):
                self.runner.train_epoch = lambda *args, **kwargs: build_model.train_epoch(
                    self.runner, *args, **kwargs)
            self.runner.model = self.model
            self.runner.optimizer = self.optimizer
            self.loss_fn = loss_fn
            self.runner.loss_fn = self.loss_fn
            self.tensor_dict_split_fn_kwargs = self.runner.tensor_dict_split_fn_kwargs
            self.initialize_tensorkeys_for_functions()
        else:
            self.lambda_opt = None
            self.loss_fn = None
            self.model = build_model()
            from .runner_generic import GenericTaskRunner
            self.runner = GenericTaskRunner(model=self.model, **kwargs)

    def __getattribute__(self, attr):
        """Direct call into self.runner methods if necessary."""
        if attr in ['reset_opt_vars', 'initialize_globals',
                    'set_tensor_dict', 'get_tensor_dict',
                    'get_required_tensorkeys_for_function',
                    'initialize_tensorkeys_for_functions',
                    'save_native', 'load_native', 'rebuild_model',
                    'set_optimizer_treatment',
                    'train', 'train_batches', 'validate']:
            return self.runner.__getattribute__(attr)
        return super(FederatedModel, self).__getattribute__(attr)

    def setup(self, num_collaborators, **kwargs):
        """
        Create new models for all of the collaborators in the experiment.

        Args:
            num_collaborators:  Number of experiment collaborators

        Returns:
            List of models
        """
        return [
            FederatedModel(
                self.build_model,
                optimizer=self.lambda_opt,
                loss_fn=self.loss_fn,
                data_loader=data_slice,
                **kwargs
            )
            for data_slice in self.data_loader.split(num_collaborators)]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Shard descriptor."""

from typing import Iterable
from typing import List

import numpy as np


class ShardDataset:
    """Shard dataset class."""

    def __len__(self) -> int:
        """Return the len of the shard dataset."""
        raise NotImplementedError

    def __getitem__(self, index: int):
        """Return an item by the index."""
        raise NotImplementedError


class ShardDescriptor:
    """Shard descriptor class."""

    def get_dataset(self, dataset_type: str) -> ShardDataset:
        """Return a shard dataset by type."""
        raise NotImplementedError

    @property
    def sample_shape(self) -> List[int]:
        """Return the sample shape info."""
        raise NotImplementedError

    @property
    def target_shape(self) -> List[int]:
        """Return the target shape info."""
        raise NotImplementedError

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return ''


class DummyShardDataset(ShardDataset):
    """Dummy shard dataset class."""

    def __init__(
            self, *,
            size: int,
            sample_shape: List[int],
            target_shape: List[int]
    ):
        """Initialize DummyShardDataset."""
        self.size = size
        self.samples = np.random.randint(0, 255, (self.size, *sample_shape), np.uint8)
        self.targets = np.random.randint(0, 255, (self.size, *target_shape), np.uint8)

    def __len__(self) -> int:
        """Return the len of the dataset."""
        return self.size

    def __getitem__(self, index: int):
        """Return a item by the index."""
        return self.samples[index], self.targets[index]


class DummyShardDescriptor(ShardDescriptor):
    """Dummy shard descriptor class."""

    def __init__(
            self,
            sample_shape: Iterable[str],
            target_shape: Iterable[str],
            size: int
    ) -> None:
        """Initialize DummyShardDescriptor."""
        self._sample_shape = [int(dim) for dim in sample_shape]
        self._target_shape = [int(dim) for dim in target_shape]
        self.size = size

    def get_dataset(self, dataset_type: str) -> ShardDataset:
        """Return a shard dataset by type."""
        return DummyShardDataset(
            size=self.size,
            sample_shape=self._sample_shape,
            target_shape=self._target_shape
        )

    @property
    def sample_shape(self) -> List[int]:
        """Return the sample shape info."""
        return self._sample_shape

    @property
    def target_shape(self) -> List[int]:
        """Return the target shape info."""
        return self._target_shape

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return 'Dummy shard descriptor'



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import tensorflow.compat.v1 as tf

from openfl.federated import TensorFlowTaskRunner

tf.disable_v2_behavior()


class TensorFlow2DUNet(TensorFlowTaskRunner):
    """Initialize.

    Args:
        **kwargs: Additional parameters to pass to the function

    """

    def __init__(self, **kwargs):
        """Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function

        """
        super().__init__(**kwargs)

        self.create_model(**kwargs)
        self.initialize_tensorkeys_for_functions()

    def create_model(self,
                     training_smoothing=32.0,
                     validation_smoothing=1.0,
                     **kwargs):
        """Create the TensorFlow 2D U-Net model.

        Args:
            training_smoothing (float): (Default=32.0)
            validation_smoothing (float): (Default=1.0)
            **kwargs: Additional parameters to pass to the function

        """
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        config.intra_op_parallelism_threads = 112
        config.inter_op_parallelism_threads = 1
        self.sess = tf.Session(config=config)

        self.X = tf.placeholder(tf.float32, self.input_shape)
        self.y = tf.placeholder(tf.float32, self.input_shape)
        self.output = define_model(self.X, use_upsampling=True, **kwargs)

        self.loss = dice_coef_loss(self.y, self.output, smooth=training_smoothing)
        self.loss_name = dice_coef_loss.__name__
        self.validation_metric = dice_coef(
            self.y, self.output, smooth=validation_smoothing)
        self.validation_metric_name = dice_coef.__name__

        self.global_step = tf.train.get_or_create_global_step()

        self.tvars = tf.trainable_variables()

        self.optimizer = tf.train.RMSPropOptimizer(1e-2)

        self.gvs = self.optimizer.compute_gradients(self.loss, self.tvars)
        self.train_step = self.optimizer.apply_gradients(self.gvs,
                                                         global_step=self.global_step)

        self.opt_vars = self.optimizer.variables()

        # FIXME: Do we really need to share the opt_vars?
        # Two opt_vars for one tvar: gradient and square sum for RMSprop.
        self.fl_vars = self.tvars + self.opt_vars

        self.initialize_globals()


def dice_coef(y_true, y_pred, smooth=1.0, **kwargs):
    """Dice coefficient.

    Calculate the Dice Coefficient

    Args:
        y_true: Ground truth annotation array
        y_pred: Prediction array from model
        smooth (float): Laplace smoothing factor (Default=1.0)
        **kwargs: Additional parameters to pass to the function

    Returns:
        float: Dice cofficient metric

    """
    intersection = tf.reduce_sum(y_true * y_pred, axis=[1, 2, 3])
    coef = (
        (tf.constant(2.) * intersection + tf.constant(smooth))
        / (tf.reduce_sum(y_true, axis=[1, 2, 3])
           + tf.reduce_sum(y_pred, axis=[1, 2, 3]) + tf.constant(smooth))
    )
    return tf.reduce_mean(coef)


def dice_coef_loss(y_true, y_pred, smooth=1.0, **kwargs):
    """Dice coefficient loss.

    Calculate the -log(Dice Coefficient) loss

    Args:
        y_true: Ground truth annotation array
        y_pred: Prediction array from model
        smooth (float): Laplace smoothing factor (Default=1.0)
        **kwargs: Additional parameters to pass to the function

    Returns:
        float: -log(Dice cofficient) metric

    """
    intersection = tf.reduce_sum(y_true * y_pred, axis=(1, 2, 3))

    term1 = -tf.log(tf.constant(2.0) * intersection + smooth)
    term2 = tf.log(tf.reduce_sum(y_true, axis=(1, 2, 3))
                   + tf.reduce_sum(y_pred, axis=(1, 2, 3)) + smooth)

    term1 = tf.reduce_mean(term1)
    term2 = tf.reduce_mean(term2)

    loss = term1 + term2

    return loss


CHANNEL_LAST = True
if CHANNEL_LAST:
    concat_axis = -1
    data_format = 'channels_last'
else:
    concat_axis = 1
    data_format = 'channels_first'

tf.keras.backend.set_image_data_format(data_format)


def define_model(input_tensor,
                 use_upsampling=False,
                 n_cl_out=1,
                 dropout=0.2,
                 print_summary=True,
                 activation_function='relu',
                 seed=0xFEEDFACE,
                 depth=5,
                 dropout_at=None,
                 initial_filters=32,
                 batch_norm=True,
                 **kwargs):
    """Define the TensorFlow model.

    Args:
        input_tensor: input shape ot the model
        use_upsampling (bool): True = use bilinear interpolation;
                               False = use transposed convolution (Default=False)
        n_cl_out (int): Number of channels in input layer (Default=1)
        dropout (float): Dropout percentage (Default=0.2)
        print_summary (bool): True = print the model summary (Default = True)
        activation_function: The activation function to use after convolutional
         layers (Default='relu')
        seed: random seed (Default=0xFEEDFACE)
        depth (int): Number of max pooling layers in encoder (Default=5)
        dropout_at: Layers to perform dropout after (Default=[2,3])
        initial_filters (int): Number of filters in first convolutional
         layer (Default=32)
        batch_norm (bool): True = use batch normalization (Default=True)
        **kwargs: Additional parameters to pass to the function

    """
    if dropout_at is None:
        dropout_at = [2, 3]
    # Set keras learning phase to train
    tf.keras.backend.set_learning_phase(True)

    # Don't initialize variables on the fly
    tf.keras.backend.manual_variable_initialization(False)

    inputs = tf.keras.layers.Input(tensor=input_tensor, name='Images')

    if activation_function == 'relu':
        activation = tf.nn.relu
    elif activation_function == 'leakyrelu':
        activation = tf.nn.leaky_relu

    params = {
        'activation': activation,
        'data_format': data_format,
        'kernel_initializer': tf.keras.initializers.he_uniform(seed=seed),
        'kernel_size': (3, 3),
        'padding': 'same',
    }

    convb_layers = {}

    net = inputs
    filters = initial_filters
    for i in range(depth):
        name = f'conv{i + 1}a'
        net = tf.keras.layers.Conv2D(name=name, filters=filters, **params)(net)
        if i in dropout_at:
            net = tf.keras.layers.Dropout(dropout)(net)
        name = f'conv{i + 1}b'
        net = tf.keras.layers.Conv2D(name=name, filters=filters, **params)(net)
        if batch_norm:
            net = tf.keras.layers.BatchNormalization()(net)
        convb_layers[name] = net
        # only pool if not last level
        if i != depth - 1:
            name = f'pool{i + 1}'
            net = tf.keras.layers.MaxPooling2D(name=name, pool_size=(2, 2))(net)
            filters *= 2

    # do the up levels
    filters //= 2
    for i in range(depth - 1):
        if use_upsampling:
            up = tf.keras.layers.UpSampling2D(
                name=f'up{depth + i + 1}', size=(2, 2))(net)
        else:
            up = tf.keras.layers.Conv2DTranspose(
                name='transConv6', filters=filters, data_format=data_format,
                kernel_size=(2, 2), strides=(2, 2), padding='same')(net)
        net = tf.keras.layers.concatenate(
            [up, convb_layers[f'conv{depth - i - 1}b']],
            axis=concat_axis
        )
        net = tf.keras.layers.Conv2D(
            name=f'conv{depth + i + 1}a',
            filters=filters, **params)(net)
        net = tf.keras.layers.Conv2D(
            name=f'conv{depth + i + 1}b',
            filters=filters, **params)(net)
        filters //= 2

    net = tf.keras.layers.Conv2D(name='Mask', filters=n_cl_out,
                                 kernel_size=(1, 1), data_format=data_format,
                                 activation='sigmoid')(net)

    model = tf.keras.models.Model(inputs=[inputs], outputs=[net])

    if print_summary:
        print(model.summary())

    return net



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/protocols/base.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bopenfl/protocols/base.proto\"+\n\nModelProto\x12\x1d\n\x07tensors\x18\x01 \x03(\x0b\x32\x0c.NamedTensor\"\xa3\x01\n\x0bNamedTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x10\n\x08lossless\x18\x03 \x01(\x08\x12\x0e\n\x06report\x18\x04 \x01(\x08\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12,\n\x14transformer_metadata\x18\x06 \x03(\x0b\x32\x0e.MetadataProto\x12\x12\n\ndata_bytes\x18\x07 \x01(\x0c\"\xac\x01\n\rMetadataProto\x12\x34\n\x0cint_to_float\x18\x01 \x03(\x0b\x32\x1e.MetadataProto.IntToFloatEntry\x12\x10\n\x08int_list\x18\x02 \x03(\x05\x12\x11\n\tbool_list\x18\x03 \x03(\x08\x12\r\n\x05model\x18\x04 \x01(\x08\x1a\x31\n\x0fIntToFloatEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\"+\n\nDataStream\x12\x0c\n\x04size\x18\x01 \x01(\r\x12\x0f\n\x07npbytes\x18\x02 \x01(\x0c\"\x81\x01\n\x17\x43ollaboratorDescription\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\x10\n\x08progress\x18\x03 \x01(\x02\x12\r\n\x05round\x18\x04 \x01(\r\x12\x14\n\x0c\x63urrent_task\x18\x05 \x01(\t\x12\x11\n\tnext_task\x18\x06 \x01(\t\"4\n\x0fTaskDescription\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\".\n\x0e\x44ownloadStatus\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\"R\n\x10\x44ownloadStatuses\x12\x1f\n\x06models\x18\x01 \x03(\x0b\x32\x0f.DownloadStatus\x12\x1d\n\x04logs\x18\x02 \x03(\x0b\x32\x0f.DownloadStatus\"\xf4\x01\n\x15\x45xperimentDescription\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\x10\n\x08progress\x18\x03 \x01(\x02\x12\x14\n\x0ctotal_rounds\x18\x04 \x01(\r\x12\x15\n\rcurrent_round\x18\x05 \x01(\r\x12,\n\x11\x64ownload_statuses\x18\x06 \x01(\x0b\x32\x11.DownloadStatuses\x12/\n\rcollaborators\x18\x07 \x03(\x0b\x32\x18.CollaboratorDescription\x12\x1f\n\x05tasks\x18\x08 \x03(\x0b\x32\x10.TaskDescriptionb\x06proto3')



_MODELPROTO = DESCRIPTOR.message_types_by_name['ModelProto']
_NAMEDTENSOR = DESCRIPTOR.message_types_by_name['NamedTensor']
_METADATAPROTO = DESCRIPTOR.message_types_by_name['MetadataProto']
_METADATAPROTO_INTTOFLOATENTRY = _METADATAPROTO.nested_types_by_name['IntToFloatEntry']
_DATASTREAM = DESCRIPTOR.message_types_by_name['DataStream']
_COLLABORATORDESCRIPTION = DESCRIPTOR.message_types_by_name['CollaboratorDescription']
_TASKDESCRIPTION = DESCRIPTOR.message_types_by_name['TaskDescription']
_DOWNLOADSTATUS = DESCRIPTOR.message_types_by_name['DownloadStatus']
_DOWNLOADSTATUSES = DESCRIPTOR.message_types_by_name['DownloadStatuses']
_EXPERIMENTDESCRIPTION = DESCRIPTOR.message_types_by_name['ExperimentDescription']
ModelProto = _reflection.GeneratedProtocolMessageType('ModelProto', (_message.Message,), {
  'DESCRIPTOR' : _MODELPROTO,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:ModelProto)
  })
_sym_db.RegisterMessage(ModelProto)

NamedTensor = _reflection.GeneratedProtocolMessageType('NamedTensor', (_message.Message,), {
  'DESCRIPTOR' : _NAMEDTENSOR,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:NamedTensor)
  })
_sym_db.RegisterMessage(NamedTensor)

MetadataProto = _reflection.GeneratedProtocolMessageType('MetadataProto', (_message.Message,), {

  'IntToFloatEntry' : _reflection.GeneratedProtocolMessageType('IntToFloatEntry', (_message.Message,), {
    'DESCRIPTOR' : _METADATAPROTO_INTTOFLOATENTRY,
    '__module__' : 'openfl.protocols.base_pb2'
    # @@protoc_insertion_point(class_scope:MetadataProto.IntToFloatEntry)
    })
  ,
  'DESCRIPTOR' : _METADATAPROTO,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:MetadataProto)
  })
_sym_db.RegisterMessage(MetadataProto)
_sym_db.RegisterMessage(MetadataProto.IntToFloatEntry)

DataStream = _reflection.GeneratedProtocolMessageType('DataStream', (_message.Message,), {
  'DESCRIPTOR' : _DATASTREAM,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:DataStream)
  })
_sym_db.RegisterMessage(DataStream)

CollaboratorDescription = _reflection.GeneratedProtocolMessageType('CollaboratorDescription', (_message.Message,), {
  'DESCRIPTOR' : _COLLABORATORDESCRIPTION,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:CollaboratorDescription)
  })
_sym_db.RegisterMessage(CollaboratorDescription)

TaskDescription = _reflection.GeneratedProtocolMessageType('TaskDescription', (_message.Message,), {
  'DESCRIPTOR' : _TASKDESCRIPTION,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:TaskDescription)
  })
_sym_db.RegisterMessage(TaskDescription)

DownloadStatus = _reflection.GeneratedProtocolMessageType('DownloadStatus', (_message.Message,), {
  'DESCRIPTOR' : _DOWNLOADSTATUS,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:DownloadStatus)
  })
_sym_db.RegisterMessage(DownloadStatus)

DownloadStatuses = _reflection.GeneratedProtocolMessageType('DownloadStatuses', (_message.Message,), {
  'DESCRIPTOR' : _DOWNLOADSTATUSES,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:DownloadStatuses)
  })
_sym_db.RegisterMessage(DownloadStatuses)

ExperimentDescription = _reflection.GeneratedProtocolMessageType('ExperimentDescription', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTDESCRIPTION,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:ExperimentDescription)
  })
_sym_db.RegisterMessage(ExperimentDescription)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _METADATAPROTO_INTTOFLOATENTRY._options = None
  _METADATAPROTO_INTTOFLOATENTRY._serialized_options = b'8\001'
  _MODELPROTO._serialized_start=31
  _MODELPROTO._serialized_end=74
  _NAMEDTENSOR._serialized_start=77
  _NAMEDTENSOR._serialized_end=240
  _METADATAPROTO._serialized_start=243
  _METADATAPROTO._serialized_end=415
  _METADATAPROTO_INTTOFLOATENTRY._serialized_start=366
  _METADATAPROTO_INTTOFLOATENTRY._serialized_end=415
  _DATASTREAM._serialized_start=417
  _DATASTREAM._serialized_end=460
  _COLLABORATORDESCRIPTION._serialized_start=463
  _COLLABORATORDESCRIPTION._serialized_end=592
  _TASKDESCRIPTION._serialized_start=594
  _TASKDESCRIPTION._serialized_end=646
  _DOWNLOADSTATUS._serialized_start=648
  _DOWNLOADSTATUS._serialized_end=694
  _DOWNLOADSTATUSES._serialized_start=696
  _DOWNLOADSTATUSES._serialized_end=778
  _EXPERIMENTDESCRIPTION._serialized_start=781
  _EXPERIMENTDESCRIPTION._serialized_end=1025
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Histology Shard Descriptor."""

import logging
import os
from pathlib import Path
from typing import Tuple
from urllib.request import urlretrieve
from zipfile import ZipFile

import numpy as np
from PIL import Image

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor
from openfl.utilities import tqdm_report_hook
from openfl.utilities import validate_file_hash


logger = logging.getLogger(__name__)


class HistologyShardDataset(ShardDataset):
    """Histology shard dataset class."""

    TRAIN_SPLIT_RATIO = 0.8

    def __init__(self, data_folder: Path, data_type='train', rank=1, worldsize=1):
        """Histology shard dataset class."""
        self.data_type = data_type
        root = Path(data_folder) / 'Kather_texture_2016_image_tiles_5000'
        classes = [d.name for d in root.iterdir() if d.is_dir()]
        class_to_idx = {cls_name: i for i, cls_name in enumerate(classes)}
        self.samples = []
        root = root.expanduser()
        for target_class in sorted(class_to_idx.keys()):
            class_index = class_to_idx[target_class]
            target_dir = os.path.join(root, target_class)
            for class_root, _, fnames in sorted(os.walk(target_dir, followlinks=True)):
                for fname in sorted(fnames):
                    path = os.path.join(class_root, fname)
                    item = path, class_index
                    self.samples.append(item)

        idx_range = list(range(len(self.samples)))
        idx_sep = int(len(idx_range) * HistologyShardDataset.TRAIN_SPLIT_RATIO)
        train_idx, test_idx = np.split(idx_range, [idx_sep])
        if data_type == 'train':
            self.idx = train_idx[rank - 1::worldsize]
        else:
            self.idx = test_idx[rank - 1::worldsize]

    def __len__(self) -> int:
        """Return the len of the shard dataset."""
        return len(self.idx)

    def load_pil(self, path):
        """Load image."""
        with open(path, 'rb') as f:
            img = Image.open(f)
            return img.convert('RGB')

    def __getitem__(self, index: int) -> Tuple['Image', int]:
        """Return an item by the index."""
        path, target = self.samples[self.idx[index]]
        sample = self.load_pil(path)
        return sample, target


class HistologyShardDescriptor(ShardDescriptor):
    """Shard descriptor class."""

    URL = ('https://zenodo.org/record/53169/files/Kather_'
           'texture_2016_image_tiles_5000.zip?download=1')
    FILENAME = 'Kather_texture_2016_image_tiles_5000.zip'
    ZIP_SHA384 = ('7d86abe1d04e68b77c055820c2a4c582a1d25d2983e38ab724e'
                  'ac75affce8b7cb2cbf5ba68848dcfd9d84005d87d6790')
    DEFAULT_PATH = Path.home() / '.openfl' / 'data'

    def __init__(
            self,
            data_folder: Path = DEFAULT_PATH,
            rank_worldsize: str = '1,1',
            **kwargs
    ):
        """Initialize HistologyShardDescriptor."""
        self.data_folder = Path.cwd() / data_folder
        self.download_data()
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))

    def download_data(self):
        """Download prepared shard dataset."""
        os.makedirs(self.data_folder, exist_ok=True)
        filepath = self.data_folder / HistologyShardDescriptor.FILENAME
        if not filepath.exists():
            reporthook = tqdm_report_hook()
            urlretrieve(HistologyShardDescriptor.URL, filepath, reporthook)  # nosec
            validate_file_hash(filepath, HistologyShardDescriptor.ZIP_SHA384)
            with ZipFile(filepath, 'r') as f:
                f.extractall(self.data_folder)

    def get_dataset(self, dataset_type):
        """Return a shard dataset by type."""
        return HistologyShardDataset(
            data_folder=self.data_folder,
            data_type=dataset_type,
            rank=self.rank,
            worldsize=self.worldsize
        )

    @property
    def sample_shape(self):
        """Return the sample shape info."""
        shape = self.get_dataset('train')[0][0].size
        return [str(dim) for dim in shape]

    @property
    def target_shape(self):
        """Return the target shape info."""
        target = self.get_dataset('train')[0][1]
        shape = np.array([target]).shape
        return [str(dim) for dim in shape]

    @property
    def dataset_description(self) -> str:
        """Return the shard dataset description."""
        return (f'Histology dataset, shard number {self.rank}'
                f' out of {self.worldsize}')



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from openfl.protocols import aggregator_pb2 as openfl_dot_protocols_dot_aggregator__pb2
from openfl.protocols import base_pb2 as openfl_dot_protocols_dot_base__pb2


class AggregatorStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetTasks = channel.unary_unary(
                '/openfl.aggregator.Aggregator/GetTasks',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksResponse.FromString,
                )
        self.GetSynch = channel.unary_unary(
                '/openfl.aggregator.Aggregator/GetSynch',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.SynchRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.SynchResponse.FromString,
                )
        self.GetTensor = channel.unary_unary(
                '/openfl.aggregator.Aggregator/GetTensor',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetTensorRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetTensorResponse.FromString,
                )
        self.SendLocalTaskResults = channel.stream_unary(
                '/openfl.aggregator.Aggregator/SendLocalTaskResults',
                request_serializer=openfl_dot_protocols_dot_base__pb2.DataStream.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.SendLocalTaskResultsResponse.FromString,
                )
        self.GetMetricStream = channel.unary_stream(
                '/openfl.aggregator.Aggregator/GetMetricStream',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetMetricStreamRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetMetricStreamResponse.FromString,
                )
        self.GetTrainedModel = channel.unary_unary(
                '/openfl.aggregator.Aggregator/GetTrainedModel',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetTrainedModelRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.TrainedModelResponse.FromString,
                )
        self.GetExperimentDescription = channel.unary_unary(
                '/openfl.aggregator.Aggregator/GetExperimentDescription',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetExperimentDescriptionRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetExperimentDescriptionResponse.FromString,
                )


class AggregatorServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetTasks(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSynch(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTensor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendLocalTaskResults(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMetricStream(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTrainedModel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetExperimentDescription(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AggregatorServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetTasks': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTasks,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksResponse.SerializeToString,
            ),
            'GetSynch': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSynch,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.SynchRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.SynchResponse.SerializeToString,
            ),
            'GetTensor': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTensor,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetTensorRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetTensorResponse.SerializeToString,
            ),
            'SendLocalTaskResults': grpc.stream_unary_rpc_method_handler(
                    servicer.SendLocalTaskResults,
                    request_deserializer=openfl_dot_protocols_dot_base__pb2.DataStream.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.SendLocalTaskResultsResponse.SerializeToString,
            ),
            'GetMetricStream': grpc.unary_stream_rpc_method_handler(
                    servicer.GetMetricStream,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetMetricStreamRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetMetricStreamResponse.SerializeToString,
            ),
            'GetTrainedModel': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTrainedModel,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetTrainedModelRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.TrainedModelResponse.SerializeToString,
            ),
            'GetExperimentDescription': grpc.unary_unary_rpc_method_handler(
                    servicer.GetExperimentDescription,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetExperimentDescriptionRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetExperimentDescriptionResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'openfl.aggregator.Aggregator', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Aggregator(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetTasks(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.aggregator.Aggregator/GetTasks',
            openfl_dot_protocols_dot_aggregator__pb2.GetTasksRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.GetTasksResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetSynch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.aggregator.Aggregator/GetSynch',
            openfl_dot_protocols_dot_aggregator__pb2.SynchRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.SynchResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTensor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.aggregator.Aggregator/GetTensor',
            openfl_dot_protocols_dot_aggregator__pb2.GetTensorRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.GetTensorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendLocalTaskResults(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/openfl.aggregator.Aggregator/SendLocalTaskResults',
            openfl_dot_protocols_dot_base__pb2.DataStream.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.SendLocalTaskResultsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetMetricStream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/openfl.aggregator.Aggregator/GetMetricStream',
            openfl_dot_protocols_dot_aggregator__pb2.GetMetricStreamRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.GetMetricStreamResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTrainedModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.aggregator.Aggregator/GetTrainedModel',
            openfl_dot_protocols_dot_aggregator__pb2.GetTrainedModelRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.TrainedModelResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetExperimentDescription(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.aggregator.Aggregator/GetExperimentDescription',
            openfl_dot_protocols_dot_aggregator__pb2.GetExperimentDescriptionRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.GetExperimentDescriptionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Collaborator module."""
import time
from enum import Enum
from logging import getLogger
from time import sleep
from typing import Tuple

import numpy as np
import wandb

from openfl.databases import TensorDB
from openfl.pipelines import NoCompressionPipeline, GenericPipeline
from openfl.pipelines import TensorCodec
from openfl.protocols import utils
from openfl.utilities import TensorKey

LOG_WANDB = True


class DevicePolicy(Enum):
    """Device assignment policy."""

    CPU_ONLY = 1

    CUDA_PREFERRED = 2


class OptTreatment(Enum):
    """Optimizer Methods.

    - RESET tells each collaborator to reset the optimizer state at the beginning
    of each round.

    - CONTINUE_LOCAL tells each collaborator to continue with the local optimizer
    state from the previous round.

    - CONTINUE_GLOBAL tells each collaborator to continue with the federally
    averaged optimizer state from the previous round.
    """

    RESET = 1
    CONTINUE_LOCAL = 2
    CONTINUE_GLOBAL = 3


class Collaborator:
    r"""The Collaborator object class.

    Args:
        collaborator_name (string): The common name for the collaborator
        aggregator_uuid: The unique id for the client
        federation_uuid: The unique id for the federation
        model: The model
        opt_treatment* (string): The optimizer state treatment (Defaults to
            "CONTINUE_GLOBAL", which is aggreagated state from previous round.)

        compression_pipeline: The compression pipeline (Defaults to None)

        num_batches_per_round (int): Number of batches per round
                                     (Defaults to None)

        delta_updates* (bool): True = Only model delta gets sent.
                               False = Whole model gets sent to collaborator.
                               Defaults to False.

        single_col_cert_common_name: (Defaults to None)

    Note:
        \* - Plan setting.
    """

    def __init__(self,
                 collaborator_name,
                 aggregator_uuid,
                 federation_uuid,
                 client,
                 task_runner,
                 task_config,
                 opt_treatment='RESET',
                 device_assignment_policy='CPU_ONLY',
                 delta_updates=False,
                 compression_pipeline=None,
                 db_store_rounds=1,
                 nn=True,
                 **kwargs):
        """Initialize."""
        self.single_col_cert_common_name = None

        self.nn = nn

        if self.single_col_cert_common_name is None:
            self.single_col_cert_common_name = ''  # for protobuf compatibility
        # we would really want this as an object

        self.collaborator_name = collaborator_name
        self.aggregator_uuid = aggregator_uuid
        self.federation_uuid = federation_uuid

        if self.nn:
            self.compression_pipeline = compression_pipeline or NoCompressionPipeline()
        else:
            self.compression_pipeline = GenericPipeline(self.nn)

        self.tensor_codec = TensorCodec(self.compression_pipeline)
        self.tensor_db = TensorDB(self.nn, agg=False)
        self.db_store_rounds = db_store_rounds

        self.task_runner = task_runner
        self.delta_updates = delta_updates

        self.client = client

        self.task_config = task_config

        self.logger = getLogger(__name__)

        if not self.nn:
            # AdaBoost.F variables
            self.adaboost_coeff = np.ones(self.task_runner.get_train_data_size())
            self.model_buffer = None
            self.errors = []

        # RESET/CONTINUE_LOCAL/CONTINUE_GLOBAL
        if hasattr(OptTreatment, opt_treatment):
            self.opt_treatment = OptTreatment[opt_treatment]
        else:
            self.logger.error(f'Unknown opt_treatment: {opt_treatment.name}.')
            raise NotImplementedError(f'Unknown opt_treatment: {opt_treatment}.')

        if hasattr(DevicePolicy, device_assignment_policy):
            self.device_assignment_policy = DevicePolicy[device_assignment_policy]
        else:
            self.logger.error('Unknown device_assignment_policy: '
                              f'{device_assignment_policy.name}.')
            raise NotImplementedError(
                f'Unknown device_assignment_policy: {device_assignment_policy}.'
            )

        self.task_runner.set_optimizer_treatment(self.opt_treatment.name)

    def set_available_devices(self, cuda: Tuple[str] = ()):
        """
        Set available CUDA devices.

        Cuda tuple contains string indeces, ('1', '3').
        """
        self.cuda_devices = cuda

    def run(self):
        """Run the collaborator."""
        if LOG_WANDB:
            wandb.init(project='AdaBoost.F', entity='gmittone', group="Krvskp",
                       config={
                           "num_clients": 10,
                           "rounds": 300,
                           "seed": 1234,
                       },
                       name=self.collaborator_name
                       )
        start_time = time.time()
        while True:
            tasks, round_number, sleep_time, time_to_quit = self.get_tasks()
            if time_to_quit:
                break
            elif sleep_time > 0:
                sleep(0.01)  # some sleep function
            else:
                self.logger.info(f'Received the following tasks: {tasks}')
                for task in tasks:
                    self.do_task(task, round_number)

                # Cleaning tensor db
                self.tensor_db.clean_up(self.db_store_rounds)

        self.logger.info('End of Federation reached. Exiting...')
        print("--- %s seconds ---" % (time.time() - start_time))
        if LOG_WANDB:
            wandb.finish()

    def run_simulation(self):
        """
        Specific function for the simulation.

        After the tasks have
        been performed for a roundquit, and then the collaborator object will
        be reinitialized after the next round
        """
        while True:
            tasks, round_number, sleep_time, time_to_quit = self.get_tasks()
            if time_to_quit:
                self.logger.info('End of Federation reached. Exiting...')
                break
            elif sleep_time > 0:
                sleep(sleep_time)  # some sleep function
            else:
                self.logger.info(f'Received the following tasks: {tasks}')
                for task in tasks:
                    self.do_task(task, round_number)
                self.logger.info(f'All tasks completed on {self.collaborator_name} '
                                 f'for round {round_number}...')
                break

    def get_tasks(self):
        """Get tasks from the aggregator."""
        # logging wait time to analyze training process
        self.logger.info('Waiting for tasks...')
        tasks, round_number, sleep_time, time_to_quit = self.client.get_tasks(
            self.collaborator_name)

        return tasks, round_number, sleep_time, time_to_quit

    def do_task(self, task, round_number):
        """Do the specified task."""
        # map this task to an actual function name and kwargs
        if hasattr(self.task_runner, 'TASK_REGISTRY') and self.nn:
            func_name = task.function_name
            task_name = task.name
            kwargs = {}
            if task.task_type == 'validate':
                if task.apply_local:
                    kwargs['apply'] = 'local'
                else:
                    kwargs['apply'] = 'global'
        else:
            if isinstance(task, str):
                task_name = task
            else:
                task_name = task.name
            func_name = self.task_config[task_name]['function']
            kwargs = self.task_config[task_name]['kwargs']

        # this would return a list of what tensors we require as TensorKeys
        required_tensorkeys_relative = self.task_runner.get_required_tensorkeys_for_function(
            func_name,
            nn=self.nn,
            **kwargs
        )

        # models actually return "relative" tensorkeys of (name, LOCAL|GLOBAL,
        # round_offset)
        # so we need to update these keys to their "absolute values"
        required_tensorkeys = []
        for tname, origin, rnd_num, report, tags in required_tensorkeys_relative:
            if origin == 'GLOBAL':
                origin = self.aggregator_uuid
            else:
                origin = self.collaborator_name

            # rnd_num is the relative round. So if rnd_num is -1, get the
            # tensor from the previous round
            required_tensorkeys.append(
                TensorKey(tname, origin, rnd_num + round_number, report, tags)
            )

        # print('Required tensorkeys = {}'.format(
        # [tk[0] for tk in required_tensorkeys]))
        input_tensor_dict = self.get_numpy_dict_for_tensorkeys(
            required_tensorkeys
        )
        # now we have whatever the model needs to do the task
        if hasattr(self.task_runner, 'TASK_REGISTRY'):
            # New interactive python API
            # New `Core` TaskRunner contains registry of tasks
            func = self.task_runner.TASK_REGISTRY[func_name]
            self.logger.info('Using Interactive Python API')

            # So far 'kwargs' contained parameters read from the plan
            # those are parameters that the eperiment owner registered for
            # the task.
            # There is another set of parameters that created on the
            # collaborator side, for instance, local processing unit identifier:s
            if (self.device_assignment_policy is DevicePolicy.CUDA_PREFERRED
                    and len(self.cuda_devices) > 0):
                kwargs['device'] = f'cuda:{self.cuda_devices[0]}'
            else:
                kwargs['device'] = 'cpu'
        else:
            # TaskRunner subclassing API
            # Tasks are defined as methods of TaskRunner
            func = getattr(self.task_runner, func_name)
            self.logger.info('Using TaskRunner subclassing API')

        # TODO: this should be generalized
        if not self.nn:
            kwargs['name'] = self.collaborator_name
            if task_name == '1_train' or task_name == '2_weak_learners_validate':
                kwargs['adaboost_coeff'] = self.adaboost_coeff

        if self.nn:
            global_output_tensor_dict, local_output_tensor_dict = func(
                col_name=self.collaborator_name,
                round_num=round_number,
                input_tensor_dict=input_tensor_dict,
                **kwargs)
        else:
            global_output_tensor_dict, local_output_tensor_dict, optional = func(
                col_name=self.collaborator_name,
                round_num=round_number,
                input_tensor_dict=input_tensor_dict,
                **kwargs)

        # TODO: this should be generalized
        if not self.nn:
            if task_name == '2_weak_learners_validate':
                self.model_buffer = input_tensor_dict['generic_model']
                self.errors = optional
            elif task_name == '3_adaboost_update':
                input_tensor_dict = input_tensor_dict['generic_model']
                alpha = input_tensor_dict[0]
                best_model = int(input_tensor_dict[1])

                self.adaboost_coeff *= np.exp(alpha * self.errors[best_model])
                self.adaboost_coeff /= np.linalg.norm(self.adaboost_coeff)

                adaboost = self.tensor_db.get_tensor_from_cache(TensorKey(
                    'generic_model',
                    self.collaborator_name,
                    round_number - 1,
                    False,
                    ('adaboost',)
                ))
                if adaboost is None:
                    adaboost = self.model_buffer.replace(self.model_buffer.get(best_model), alpha)
                else:
                    adaboost.add(self.model_buffer.get(best_model), alpha)

                self.tensor_db.cache_tensor({TensorKey(
                    'generic_model',
                    self.collaborator_name,
                    round_number,
                    False,
                    ('adaboost',)
                ): adaboost})

        # Save global and local output_tensor_dicts to TensorDB
        self.tensor_db.cache_tensor(global_output_tensor_dict)
        self.tensor_db.cache_tensor(local_output_tensor_dict)

        # send the results for this tasks; delta and compression will occur in
        # this function
        self.send_task_results(global_output_tensor_dict, round_number, task_name)

        # Synchronisation point for AdaBoost.F
        # TODO: this should be applied only if the next task is GLOBAL
        if not self.nn:
            while not self.synch(task_name, round_number, self.collaborator_name):
                sleep(1)

    def synch(self, task_name, round_number, collaborator_name):
        self.logger.info('Waiting for global task completion...')
        return self.client.synch(task_name, round_number, collaborator_name)

    def get_numpy_dict_for_tensorkeys(self, tensor_keys):
        """Get tensor dictionary for specified tensorkey set."""
        return {k.tensor_name: self.get_data_for_tensorkey(k) for k in tensor_keys}

    def get_data_for_tensorkey(self, tensor_key):
        """
        Resolve the tensor corresponding to the requested tensorkey.

        Args
        ----
        tensor_key:         Tensorkey that will be resolved locally or
                            remotely. May be the product of other tensors
        """
        # try to get from the store
        tensor_name, origin, round_number, report, tags = tensor_key

        self.logger.debug(f'Attempting to retrieve tensor {tensor_key} from local store')
        # nparray = self.tensor_db.get_tensor_from_cache(tensor_key)

        nparray = None
        if 'model' in tags:
            # Pulling the model for the first time
            nparray = self.get_tensor_from_aggregator(
                tensor_key,
                require_lossless=True,
                aggregated=True
            )
        # These tags are specific to AdaBoost.F
        elif 'weak_learner' in tags:
            nparray = self.get_tensor_from_aggregator(
                tensor_key,
                require_lossless=True,
                aggregated=False,
            )
        elif 'adaboost_coeff' in tags:
            nparray = self.get_tensor_from_aggregator(
                tensor_key,
                require_lossless=True,
                aggregated=False,
            )

        if origin == self.collaborator_name:
            self.logger.debug(f'Attempting to retrieve tensor {tensor_key} from local store')
            nparray = self.tensor_db.get_tensor_from_cache(tensor_key)

            # if None and origin is our client, request it from the client
            if nparray is None:
                if origin == self.collaborator_name:
                    self.logger.info(
                        f'Attempting to find locally stored {tensor_name} tensor from prior round...'
                    )
                    prior_round = round_number - 1
                    while prior_round >= 0:
                        nparray = self.tensor_db.get_tensor_from_cache(
                            TensorKey(tensor_name, origin, prior_round, report, tags))
                        if nparray is not None:
                            self.logger.debug(f'Found tensor {tensor_name} in local TensorDB '
                                              f'for round {prior_round}')
                            return nparray
                        prior_round -= 1
                    self.logger.info(
                        f'Cannot find any prior version of tensor {tensor_name} locally...'
                    )
                    self.logger.debug('Unable to get tensor from local store...'
                                      'attempting to retrieve from client')
        else:
            # Determine whether there are additional compression related
            # dependencies.
            # Typically, dependencies are only relevant to model layers
            tensor_dependencies = self.tensor_codec.find_dependencies(
                tensor_key, self.delta_updates
            )
            if len(tensor_dependencies) > 0:
                # Resolve dependencies
                # tensor_dependencies[0] corresponds to the prior version
                # of the model.
                # If it exists locally, should pull the remote delta because
                # this is the least costly path
                prior_model_layer = self.tensor_db.get_tensor_from_cache(
                    tensor_dependencies[0]
                )
                if prior_model_layer is not None:
                    uncompressed_delta = self.get_tensor_from_aggregator(
                        tensor_dependencies[1],
                        aggregated=True
                    )
                    new_model_tk, nparray = self.tensor_codec.apply_delta(
                        tensor_dependencies[1],
                        uncompressed_delta,
                        prior_model_layer,
                        creates_model=True,
                    )
                    self.tensor_db.cache_tensor({new_model_tk: nparray})
                else:
                    self.logger.info('Count not find previous model layer.'
                                     'Fetching latest layer from aggregator')
                    # The original model tensor should be fetched from client
                    nparray = self.get_tensor_from_aggregator(
                        tensor_key,
                        require_lossless=True,
                        aggregated=True
                    )

        return nparray

    def get_tensor_from_aggregator(self, tensor_key,
                                   require_lossless=False, aggregated=True):
        """
        Return the decompressed tensor associated with the requested tensor key.

        If the key requests a compressed tensor (in the tag), the tensor will
        be decompressed before returning
        If the key specifies an uncompressed tensor (or just omits a compressed
        tag), the decompression operation will be skipped

        Args
        ----
        tensor_key  :               The requested tensor
        require_lossless:   Should compression of the tensor be allowed
                                    in flight?
                                    For the initial model, it may affect
                                    convergence to apply lossy
                                    compression. And metrics shouldn't be
                                    compressed either

        Returns
        -------
        nparray     : The decompressed tensor associated with the requested
                      tensor key
        """
        tensor_name, origin, round_number, report, tags = tensor_key

        self.logger.debug(f'Requesting tensor {tensor_key}')
        tensor = self.client.get_tensor(
            self.collaborator_name, tensor_name, round_number, report, tags, require_lossless, aggregated)

        # this translates to a numpy array and includes decompression, as
        # necessary
        nparray = self.named_tensor_to_nparray(tensor)

        # cache this tensor
        self.tensor_db.cache_tensor({tensor_key: nparray})

        return nparray

    def send_task_results(self, tensor_dict, round_number, task_name):
        """Send task results to the aggregator."""
        named_tensors = [
            self.nparray_to_named_tensor(k, v) for k, v in tensor_dict.items()
        ]

        # for general tasks, there may be no notion of data size to send.
        # But that raises the question how to properly aggregate results.

        data_size = -1

        if 'train' in task_name:
            data_size = self.task_runner.get_train_data_size()

        if 'valid' in task_name:
            data_size = self.task_runner.get_valid_data_size()

        self.logger.debug(f'{task_name} data size = {data_size}')

        for tensor in tensor_dict:
            tensor_name, origin, fl_round, report, tags = tensor

            if report:
                self.logger.metric(
                    f'Round {round_number}, collaborator {self.collaborator_name} '
                    f'is sending metric for task {task_name}:'
                    f' {tensor_name}\t{tensor_dict[tensor]}')

        self.client.send_local_task_results(
            self.collaborator_name, round_number, task_name, data_size, named_tensors)

    def nparray_to_named_tensor(self, tensor_key, nparray):
        """
        Construct the NamedTensor Protobuf.

        Includes logic to create delta, compress tensors with the TensorCodec, etc.
        """
        # if we have an aggregated tensor, we can make a delta
        tensor_name, origin, round_number, report, tags = tensor_key
        if 'trained' in tags and self.delta_updates:
            # Should get the pretrained model to create the delta. If training
            # has happened,
            # Model should already be stored in the TensorDB
            model_nparray = self.tensor_db.get_tensor_from_cache(
                TensorKey(
                    tensor_name,
                    origin,
                    round_number,
                    report,
                    ('model',)
                )
            )

            # The original model will not be present for the optimizer on the
            # first round.
            if model_nparray is not None:
                delta_tensor_key, delta_nparray = self.tensor_codec.generate_delta(
                    tensor_key,
                    nparray,
                    model_nparray
                )
                delta_comp_tensor_key, delta_comp_nparray, metadata = self.tensor_codec.compress(
                    delta_tensor_key,
                    delta_nparray
                )

                named_tensor = utils.construct_named_tensor(
                    delta_comp_tensor_key,
                    delta_comp_nparray,
                    metadata,
                    lossless=False
                )
                return named_tensor

        # Assume every other tensor requires lossless compression
        compressed_tensor_key, compressed_nparray, metadata = self.tensor_codec.compress(
            tensor_key,
            nparray,
            require_lossless=True
        )
        named_tensor = utils.construct_named_tensor(
            compressed_tensor_key,
            compressed_nparray,
            metadata,
            lossless=True
        )

        return named_tensor

    def named_tensor_to_nparray(self, named_tensor):
        """Convert named tensor to a numpy array."""
        # do the stuff we do now for decompression and frombuffer and stuff
        # This should probably be moved back to protoutils
        raw_bytes = named_tensor.data_bytes
        if self.nn:
            metadata = [{'int_to_float': proto.int_to_float,
                         'int_list': proto.int_list,
                         'bool_list': proto.bool_list
                         } for proto in named_tensor.transformer_metadata]
        else:
            metadata = [{'int_to_float': proto.int_to_float,
                         'int_list': proto.int_list,
                         'bool_list': proto.bool_list,
                         'model': proto.model,
                         } for proto in named_tensor.transformer_metadata]
        # The tensor has already been transfered to collaborator, so
        # the newly constructed tensor should have the collaborator origin
        tensor_key = TensorKey(
            named_tensor.name,
            self.collaborator_name,
            named_tensor.round_number,
            named_tensor.report,
            tuple(named_tensor.tags)
        )
        tensor_name, origin, round_number, report, tags = tensor_key
        if 'compressed' in tags:
            decompressed_tensor_key, decompressed_nparray = self.tensor_codec.decompress(
                tensor_key,
                data=raw_bytes,
                transformer_metadata=metadata,
                require_lossless=True
            )
        elif 'lossy_compressed' in tags:
            decompressed_tensor_key, decompressed_nparray = self.tensor_codec.decompress(
                tensor_key,
                data=raw_bytes,
                transformer_metadata=metadata
            )
        else:
            # There could be a case where the compression pipeline is bypassed
            # entirely
            self.logger.warning('Bypassing tensor codec...')
            decompressed_tensor_key = tensor_key
            decompressed_nparray = raw_bytes

        self.tensor_db.cache_tensor(
            {decompressed_tensor_key: decompressed_nparray}
        )

        return decompressed_nparray



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Keras optimizers package."""
import pkgutil

if pkgutil.find_loader('tensorflow'):
    from .fedprox import FedProxOptimizer # NOQA



"""GenericPipeline module."""

from .pipeline import TransformationPipeline, Float32NumpyArrayToBytes


class GenericPipeline(TransformationPipeline):
    """The data pipeline without any compression."""

    def __init__(self, nn=True, **kwargs):
        """Initialize."""
        super(GenericPipeline, self).__init__(transformers=[Float32NumpyArrayToBytes(nn)], nn=nn,  **kwargs)
        self.nn = nn

    def is_nn(self):
        return self.nn